 /* 
  * UAE - The Un*x Amiga Emulator
  * 
  * Memory access functions
  *
  * Copyright 1996 Bernd Schmidt
  */

#ifndef BIGENDIAN
#ifndef ASM_MEMORY

static __inline__ ULONG do_get_mem_long(ULONG *a)
{
    UBYTE *b = (UBYTE *)a;
    
    return (*b << 24) | (*(b+1) << 16) | (*(b+2) << 8) | (*(b+3));
}

static __inline__ UWORD do_get_mem_word(UWORD *a)
{
    UBYTE *b = (UBYTE *)a;
    
    return (*b << 8) | (*(b+1));
}

static __inline__ UBYTE do_get_mem_byte(UBYTE *a)
{
    return *a;
}

static __inline__ void do_put_mem_long(ULONG *a, ULONG v)
{
    UBYTE *b = (UBYTE *)a;
    
    *b = v >> 24;
    *(b+1) = v >> 16;    
    *(b+2) = v >> 8;
    *(b+3) = v;
}

static __inline__ void do_put_mem_word(UWORD *a, UWORD v)
{
    UBYTE *b = (UBYTE *)a;
    
    *b = v >> 8;
    *(b+1) = v;
}

static __inline__ void do_put_mem_byte(UBYTE *a, UBYTE v)
{
    *a = v;
}

#define call_mem_get_func(func, addr) ((*func)(addr))
#define call_mem_put_func(func, addr, v) ((*func)(addr, v))

#undef NO_INLINE_MEMORY_ACCESS
#undef MD_HAVE_MEM_1_FUNCS

#else /* ASM_MEMORY */

static __inline__ ULONG do_get_mem_long(ULONG *a)
{
    ULONG retval;

    __asm__ ("bswapl %0" : "=r" (retval) : "0" (*a) : "cc");
    return retval;
}

static __inline__ ULONG do_get_mem_word(UWORD *a)
{
    ULONG retval;

    __asm__ ("xorl %k0,%k0\n\tmovw %w1,%w0\n\trolw $8,%w0" : "=&r" (retval) : "m" (*a) : "cc");
    return retval;
}

#define do_get_mem_byte(a) ((ULONG)*(a))

static __inline__ void do_put_mem_long(ULONG *a, ULONG v)
{
    __asm__ ("bswapl %0" : "=r" (v) : "0" (v) : "cc");
    *a = v;
}

static __inline__ void do_put_mem_word(UWORD *a, UWORD v)
{
    __asm__ ("rolw $8,%0" : "=r" (v) : "0" (v) : "cc");
    *a = v;
}

#define do_put_mem_byte(a,v) (*(a) = (v))

#define call_mem_get_func(func,addr) ((*func)(addr))
#define call_mem_put_func(func,addr,v) ((*func)(addr,v))

#undef NO_INLINE_MEMORY_ACCESS
#undef MD_HAVE_MEM_1_FUNCS
#endif

#else /* Big endian memory functions */

#ifdef PPC
static __inline__ ULONG do_get_mem_long(ULONG *a)
{
    return *a;
}
#else /* fixes IRIX, Solaris, and HP-UX. */
static __inline__ ULONG do_get_mem_long(ULONG *a)
{
    UBYTE *b = (UBYTE *)a;
   
    return (*b << 24) | (*(b+1) << 16) | (*(b+2) << 8) | (*(b+3));
}
#endif

static __inline__ UWORD do_get_mem_word(UWORD *a)
{
    return *a;
}

static __inline__ UBYTE do_get_mem_byte(UBYTE *a)
{
    return *a;
}

#ifdef PPC
static __inline__ void do_put_mem_long(ULONG *a, ULONG v)
{
    *a = v;
}
#else /* fixes Solaris, IRIX, and HP-UX */
static __inline__ void do_put_mem_long(ULONG *a, ULONG v)
{
    UBYTE *b = (UBYTE *)a;
 
    *b = v >> 24;
    *(b+1) = v >> 16;
    *(b+2) = v >> 8;
    *(b+3) = v;
}  
#endif

static __inline__ void do_put_mem_word(UWORD *a, UWORD v)
{
    *a = v;
}

static __inline__ void do_put_mem_byte(UBYTE *a, UBYTE v)
{
    *a = v;
}

#define call_mem_get_func(func, addr) ((*func)(addr))
#define call_mem_put_func(func, addr, v) ((*func)(addr, v))

#undef NO_INLINE_MEMORY_ACCESS
#undef MD_HAVE_MEM_1_FUNCS

#endif

//IRIX CC needs a little new-line at the end of include files.

