 /*
  * UAE - The Un*x Amiga Emulator
  *
  * Memory management
  * * (c) 1995 Bernd Schmidt
  */

#include "sysconfig.h"
#include "sysdeps.h"

#include "config.h"
#include "options.h"

#include "custom.h"

#include "../CPU/osdep/memory.h"
#include "memory.h"

#include "readcpu.h"
#include "newcpu.h"

#ifdef _Memory_Debug
#include <stdio.h>
#endif

#include "Memory.h"
#include "IWM.h"
#include "RAM.h"
#include "ROM.h"
#include "SCC.h"
#include "SCSI.h"
#include "VIA.h"

#include "prefs.h"
#include "../Unix/universal.h"

#ifdef USE_MAPPED_MEMORY
#include <sys/mman.h>
#endif

UBYTE vOverlay = 1;

#define RAMmem_start			( (ULONG) kRAM_Base)
#define RAMmem_size			( (ULONG) kRAM_Size)

#define Overlay_RAMmem_start		( (ULONG) kRAM_Overlay_Base)
#define Overlay_RAMmem_size		( (ULONG) kRAM_Overlay_Top - \
						(ULONG) kRAM_Overlay_Base)
#define ROMmem_start			( (ULONG) kROM_Base)
#define ROMmem_size			( (ULONG) kROM_Size)

#define Overlay_ROMmem_start		( (ULONG) kROM_Overlay_Base)
#define Overlay_ROMmem_size		( (ULONG) kROM_Overlay_Top - \
						(ULONG) kROM_Overlay_Base)

int buserr;
#ifdef SAVE_MEMORY_BANKS

addrbank **mem_banks;
addrbank *Overlay_mem_banks[65536];
addrbank *Non_overlay_mem_banks[65536];

#else

addrbank *mem_banks;

#endif

#ifdef NO_INLINE_MEMORY_ACCESS
__inline__ ULONG longget(CPTR addr)
{
    return call_mem_get_func(get_mem_bank(addr).lget, addr);
}
__inline__ ULONG wordget(CPTR addr)
{
    return call_mem_get_func(get_mem_bank(addr).wget, addr);
}
__inline__ ULONG byteget(CPTR addr) 
{
    return call_mem_get_func(get_mem_bank(addr).bget, addr);
}
__inline__ void longput(CPTR addr, ULONG l)
{
    call_mem_put_func(get_mem_bank(addr).lput, addr, l);
}
__inline__ void wordput(CPTR addr, ULONG w)
{
    call_mem_put_func(get_mem_bank(addr).wput, addr, w);
}
__inline__ void byteput(CPTR addr, ULONG b)
{
    call_mem_put_func(get_mem_bank(addr).bput, addr, b);
}
#endif

/* A dummy bank that only contains zeros */

static ULONG dummy_lget(CPTR) REGPARAM;
static ULONG dummy_wget(CPTR) REGPARAM;
static ULONG dummy_bget(CPTR) REGPARAM;
static void  dummy_lput(CPTR, ULONG) REGPARAM;
static void  dummy_wput(CPTR, ULONG) REGPARAM;
static void  dummy_bput(CPTR, ULONG) REGPARAM;
static int   dummy_check(CPTR addr, ULONG size) REGPARAM;

ULONG REGPARAM2 dummy_lget(CPTR addr)
{
    return 0;
}

ULONG REGPARAM2 dummy_wget(CPTR addr)
{
    return 0;
}

ULONG REGPARAM2 dummy_bget(CPTR addr)
{
    return 0;
}

void REGPARAM2 dummy_lput(CPTR addr, ULONG l)
{
//	;
}
void REGPARAM2 dummy_wput(CPTR addr, ULONG w)
{
//	;
}
void REGPARAM2 dummy_bput(CPTR addr, ULONG b)
{
//	;
}

int REGPARAM2 dummy_check(CPTR addr, ULONG size)
{
    return 0;
}

ULONG RAMmem_mask,
	ROMmem_mask,
	Overlay_RAMmem_mask,
	Overlay_ROMmem_mask;

/* overlay RAM memory */

UBYTE *Overlay_RAMmemory;

static ULONG	Overlay_RAMmem_lget(CPTR) REGPARAM;
static ULONG	Overlay_RAMmem_wget(CPTR) REGPARAM;
static ULONG	Overlay_RAMmem_bget(CPTR) REGPARAM;
static void	Overlay_RAMmem_lput(CPTR, ULONG) REGPARAM;
static void	Overlay_RAMmem_wput(CPTR, ULONG) REGPARAM;
static void	Overlay_RAMmem_bput(CPTR, ULONG) REGPARAM;

static int	Overlay_RAMmem_check(CPTR addr, ULONG size) REGPARAM;
static UBYTE  *Overlay_RAMmem_xlate(CPTR addr) REGPARAM;

ULONG REGPARAM2 Overlay_RAMmem_lget(CPTR addr)
{
    ULONG *m;
    addr -= Overlay_RAMmem_start & Overlay_RAMmem_mask;
    addr &= Overlay_RAMmem_mask;
    m = (ULONG *)(Overlay_RAMmemory + addr);
    return do_get_mem_long(m);
}

ULONG REGPARAM2 Overlay_RAMmem_wget(CPTR addr)
{
    UWORD *m;
    addr -= Overlay_RAMmem_start & Overlay_RAMmem_mask;
    addr &= Overlay_RAMmem_mask;
    m = (UWORD *)(Overlay_RAMmemory + addr);
    return do_get_mem_word(m);
}

ULONG REGPARAM2 Overlay_RAMmem_bget(CPTR addr)
{
    addr -= Overlay_RAMmem_start & Overlay_RAMmem_mask;
    addr &= Overlay_RAMmem_mask;
    return Overlay_RAMmemory[addr];
}

void REGPARAM2 Overlay_RAMmem_lput(CPTR addr, ULONG l)
{
    ULONG *m;
    addr -= Overlay_RAMmem_start & Overlay_RAMmem_mask;
    addr &= Overlay_RAMmem_mask;
    m = (ULONG *)(Overlay_RAMmemory + addr);
    do_put_mem_long(m, l);
}

void REGPARAM2 Overlay_RAMmem_wput(CPTR addr, ULONG w)
{
    UWORD *m;
    addr -= Overlay_RAMmem_start & Overlay_RAMmem_mask;
    addr &= Overlay_RAMmem_mask;
    m = (UWORD *)(Overlay_RAMmemory + addr);
    do_put_mem_word(m, w);
}

void REGPARAM2 Overlay_RAMmem_bput(CPTR addr, ULONG b)
{
    addr -= Overlay_RAMmem_start & Overlay_RAMmem_mask;
    addr &= Overlay_RAMmem_mask;
    Overlay_RAMmemory[addr] = b;
}

int REGPARAM2 Overlay_RAMmem_check(CPTR addr, ULONG size)
{
    addr -= Overlay_RAMmem_start & Overlay_RAMmem_mask;
    addr &= Overlay_RAMmem_mask;
    return (addr + size) < Overlay_RAMmem_size;
}

UBYTE REGPARAM2 *Overlay_RAMmem_xlate(CPTR addr)
{
    addr -= Overlay_RAMmem_start & Overlay_RAMmem_mask;
    addr &= Overlay_RAMmem_mask;
    return Overlay_RAMmemory + addr;
}


/* Chip memory */

UBYTE *RAMmemory;

static ULONG RAMmem_lget(CPTR) REGPARAM;
static ULONG RAMmem_wget(CPTR) REGPARAM;
static ULONG RAMmem_bget(CPTR) REGPARAM;
static void  RAMmem_lput(CPTR, ULONG) REGPARAM;
static void  RAMmem_wput(CPTR, ULONG) REGPARAM;
static void  RAMmem_bput(CPTR, ULONG) REGPARAM;

static int   RAMmem_check(CPTR addr, ULONG size) REGPARAM;
static UBYTE *RAMmem_xlate(CPTR addr) REGPARAM;

ULONG REGPARAM2 RAMmem_lget(CPTR addr)
{
    ULONG *m;
    addr -= RAMmem_start & RAMmem_mask;
    addr &= RAMmem_mask;
    m = (ULONG *)(RAMmemory + addr);
    return do_get_mem_long(m);
}

ULONG REGPARAM2 RAMmem_wget(CPTR addr)
{
    UWORD *m;
    addr -= RAMmem_start & RAMmem_mask;
    addr &= RAMmem_mask;
    m = (UWORD *)(RAMmemory + addr);
    return do_get_mem_word(m);
}

ULONG REGPARAM2 RAMmem_bget(CPTR addr)
{
    addr -= RAMmem_start & RAMmem_mask;
    addr &= RAMmem_mask;
    return RAMmemory[addr];
}

void REGPARAM2 RAMmem_lput(CPTR addr, ULONG l)
{
    ULONG *m;
    addr -= RAMmem_start & RAMmem_mask;
    addr &= RAMmem_mask;
    m = (ULONG *)(RAMmemory + addr);
    do_put_mem_long(m, l);
}

void REGPARAM2 RAMmem_wput(CPTR addr, ULONG w)
{
    UWORD *m;
    addr -= RAMmem_start & RAMmem_mask;
    addr &= RAMmem_mask;
    m = (UWORD *)(RAMmemory + addr);
    do_put_mem_word(m, w);
}

void REGPARAM2 RAMmem_bput(CPTR addr, ULONG b)
{
    addr -= RAMmem_start & RAMmem_mask;
    addr &= RAMmem_mask;
    RAMmemory[addr] = b;
}

int REGPARAM2 RAMmem_check(CPTR addr, ULONG size)
{
    addr -= RAMmem_start & RAMmem_mask;
    addr &= RAMmem_mask;
    return (addr + size) < RAMmem_size;
}

UBYTE REGPARAM2 *RAMmem_xlate(CPTR addr)
{
    addr -= RAMmem_start & RAMmem_mask;
    addr &= RAMmem_mask;
    return RAMmemory + addr;
}



/* Overlay ROM memory */

UBYTE *Overlay_ROMmemory;

static ULONG	Overlay_ROMmem_lget(CPTR) REGPARAM;
static ULONG	Overlay_ROMmem_wget(CPTR) REGPARAM;
static ULONG	Overlay_ROMmem_bget(CPTR) REGPARAM;
static void	Overlay_ROMmem_lput(CPTR, ULONG) REGPARAM;
static void	Overlay_ROMmem_wput(CPTR, ULONG) REGPARAM;
static void	Overlay_ROMmem_bput(CPTR, ULONG) REGPARAM;
static int	Overlay_ROMmem_check(CPTR addr, ULONG size) REGPARAM;
static UBYTE  *Overlay_ROMmem_xlate(CPTR addr) REGPARAM;

ULONG REGPARAM2 Overlay_ROMmem_lget(CPTR addr)
{
    ULONG *m;
    addr -= Overlay_ROMmem_start & Overlay_ROMmem_mask;
    addr &= Overlay_ROMmem_mask;
    m = (ULONG *)(Overlay_ROMmemory + addr);
    return do_get_mem_long(m); 
}

ULONG REGPARAM2 Overlay_ROMmem_wget(CPTR addr)
{
    UWORD *m;
    addr -= Overlay_ROMmem_start & Overlay_ROMmem_mask;
    addr &= Overlay_ROMmem_mask;
    m = (UWORD *)(Overlay_ROMmemory + addr);
    return do_get_mem_word(m); 
}

ULONG REGPARAM2 Overlay_ROMmem_bget(CPTR addr)
{
    addr -= Overlay_ROMmem_start & Overlay_ROMmem_mask;
    addr &= Overlay_ROMmem_mask;
    return Overlay_ROMmemory[addr];
}

void REGPARAM2 Overlay_ROMmem_lput(CPTR addr, ULONG b)
{
//	;
}

void REGPARAM2 Overlay_ROMmem_wput(CPTR addr, ULONG b)
{
//	;
}

void REGPARAM2 Overlay_ROMmem_bput(CPTR addr, ULONG b)
{
//	;
}

int REGPARAM2 Overlay_ROMmem_check(CPTR addr, ULONG size)
{
    addr -= Overlay_ROMmem_start & Overlay_ROMmem_mask;
    addr &= Overlay_ROMmem_mask;
    return (addr + size) < Overlay_ROMmem_size;
}

UBYTE REGPARAM2 *Overlay_ROMmem_xlate(CPTR addr)
{
    addr -= Overlay_ROMmem_start & Overlay_ROMmem_mask;
    addr &= Overlay_ROMmem_mask;
    return Overlay_ROMmemory + addr;
}

/* ROM memory */

UBYTE *ROMmemory;
unsigned int ROMmem_write_enable;

static ULONG	ROMmem_lget(CPTR) REGPARAM;
static ULONG	ROMmem_wget(CPTR) REGPARAM;
static ULONG	ROMmem_bget(CPTR) REGPARAM;
static void	ROMmem_lput(CPTR, ULONG) REGPARAM;
static void	ROMmem_wput(CPTR, ULONG) REGPARAM;
static void	ROMmem_bput(CPTR, ULONG) REGPARAM;
static int	ROMmem_check(CPTR addr, ULONG size) REGPARAM;
static UBYTE	*ROMmem_xlate(CPTR addr) REGPARAM;

ULONG REGPARAM2 ROMmem_lget(CPTR addr)
{
    ULONG *m;
    addr -= ROMmem_start & ROMmem_mask;
    addr &= ROMmem_mask;
    m = (ULONG *)(ROMmemory + addr);
    return do_get_mem_long(m); 
}

ULONG REGPARAM2 ROMmem_wget(CPTR addr)
{
    UWORD *m;
    addr -= ROMmem_start & ROMmem_mask;
    addr &= ROMmem_mask;
    m = (UWORD *)(ROMmemory + addr);
    return do_get_mem_word(m); 
}

ULONG REGPARAM2 ROMmem_bget(CPTR addr)
{
    addr -= ROMmem_start & ROMmem_mask;
    addr &= ROMmem_mask;
    return ROMmemory[addr];
}

void REGPARAM2 ROMmem_lput(CPTR addr, ULONG l)
{
    ULONG *m;

    if( ROMmem_write_enable )
    {
	addr -= ROMmem_start & ROMmem_mask;
	addr &= ROMmem_mask;
	m = (ULONG *)(ROMmemory + addr);
	do_put_mem_long(m, l);
    }
}

void REGPARAM2 ROMmem_wput(CPTR addr, ULONG w)
{
    UWORD *m;

    if( ROMmem_write_enable )
    {
	addr -= ROMmem_start & ROMmem_mask;
	addr &= ROMmem_mask;
	m = (UWORD *)(ROMmemory + addr);
	do_put_mem_word(m, w);
    }
}

void REGPARAM2 ROMmem_bput(CPTR addr, ULONG b)
{

    if( ROMmem_write_enable )
    {
	addr -= ROMmem_start & ROMmem_mask;
	addr &= ROMmem_mask;
	ROMmemory[addr] = b;
    }
}

int REGPARAM2 ROMmem_check(CPTR addr, ULONG size)
{
    addr -= ROMmem_start & ROMmem_mask;
    addr &= ROMmem_mask;
    return (addr + size) < ROMmem_size;
}

UBYTE REGPARAM2 *ROMmem_xlate(CPTR addr)
{
    addr -= ROMmem_start & ROMmem_mask;
    addr &= ROMmem_mask;
    return ROMmemory + addr;
}

/* Default memory access functions */

int REGPARAM2 default_check(CPTR a, ULONG b)
{
    return 0;
}

void Mac_reset(void);
UBYTE REGPARAM2 *default_xlate(CPTR a)
{
    Mac_reset();
    return RAMmem_xlate(0); /* So we don't crash. */
}

void Mac_reset()
{
}

static ULONG devicemem_lget(CPTR) REGPARAM;
static ULONG devicemem_wget(CPTR) REGPARAM;
static ULONG devicemem_bget(CPTR) REGPARAM;
static void  devicemem_lput(CPTR, ULONG) REGPARAM;
static void  devicemem_wput(CPTR, ULONG) REGPARAM;
static void  devicemem_bput(CPTR, ULONG) REGPARAM;
static int   devicemem_check(CPTR addr, ULONG size) REGPARAM;

ULONG REGPARAM2 devicemem_lget(CPTR addr)
{
	return (ULONG) Get_MM(addr & kMemory_Extent, Size_Long);
}

ULONG REGPARAM2 devicemem_wget(CPTR addr)
{
	return (UWORD) Get_MM(addr & kMemory_Extent, Size_Word);
}

ULONG REGPARAM2 devicemem_bget(CPTR addr)
{
	return (UBYTE) Get_MM(addr & kMemory_Extent, Size_Byte);
}

void REGPARAM2 devicemem_lput(CPTR addr, ULONG l)
{
	Put_MM(addr & kMemory_Extent, Size_Long, (ULONG) l);
}

void REGPARAM2 devicemem_wput(CPTR addr, ULONG w)
{
	Put_MM(addr & kMemory_Extent, Size_Word, (ULONG) w);
}

void REGPARAM2 devicemem_bput(CPTR addr, ULONG b)
{
	Put_MM(addr & kMemory_Extent, Size_Byte, (ULONG) b);
}

/* Address banks */

addrbank dummy_bank = {
    dummy_lget, dummy_wget, dummy_bget,
    dummy_lput, dummy_wput, dummy_bput,
    default_xlate, dummy_check
};

addrbank Overlay_RAMmem_bank = {
    Overlay_RAMmem_lget, Overlay_RAMmem_wget, Overlay_RAMmem_bget,
    Overlay_RAMmem_lput, Overlay_RAMmem_wput, Overlay_RAMmem_bput,
    Overlay_RAMmem_xlate, Overlay_RAMmem_check
};

addrbank Overlay_ROMmem_bank = {
    Overlay_ROMmem_lget, Overlay_ROMmem_wget, Overlay_ROMmem_bget,
    Overlay_ROMmem_lput, Overlay_ROMmem_wput, Overlay_ROMmem_bput,
    Overlay_ROMmem_xlate, Overlay_ROMmem_check
};

addrbank RAMmem_bank = {
    RAMmem_lget, RAMmem_wget, RAMmem_bget,
    RAMmem_lput, RAMmem_wput, RAMmem_bput,
    RAMmem_xlate, RAMmem_check };


addrbank ROMmem_bank = {
    ROMmem_lget, ROMmem_wget, ROMmem_bget,
    ROMmem_lput, ROMmem_wput, ROMmem_bput,
    ROMmem_xlate, ROMmem_check
};

addrbank devicemem_bank = {
    devicemem_lget, devicemem_wget, devicemem_bget,
    devicemem_lput, devicemem_wput, devicemem_bput,
    default_xlate, default_check
};



char *address_space, *good_address_map;
int good_address_fd;
#define MAKE_USER_PROGRAMS_BEHAVE 1

void map_banks(addrbank *bank, int start, int size)
{
    int bnr;
    int hioffs = 0, endhioffs = 1;
    endhioffs = 256;

    for (hioffs = 0; hioffs < endhioffs; hioffs++)
    {
	for (bnr = start; bnr < start+size; bnr++)
	    put_mem_bank((bnr + hioffs * 256) << 16, bank);
    }
}

int Memory_Init (void)
{
	memory_init_uae_069();
	VIA_PORA4(1);

	return 0;
}

void  Memory_Reset (void)
{
	VIA_PORA4(1);
}

ULONG Get_MM (CPTR addr, Size_Ty Size)
{
  if (vOverlay)
  {
    switch (addr >> 22)
    {
      case 0 : // ROM
        if (addr < kROM_Overlay_Top)
        {
          switch (Size)
          {
            case Size_Byte : return (UBYTE) ROM_GetB (addr % kROM_Size); break;
            case Size_Word : return (UWORD) ROM_GetW (addr % kROM_Size); break;
            case Size_Long : return (ULONG) ROM_GetL (addr % kROM_Size); break;
          }
        }
		#ifdef _Memory_Debug
        switch (Size)
        {
          case Size_Byte : printf("Byte Access at %8lx\n", addr); break;
          case Size_Word : printf("Word Access at %8lx\n", addr); break;
          case Size_Long : printf("Long Access at %8lx\n", addr); break;
        }
		#endif
        return 0;
        break;
      case 1 : // ROM, SCSI and RAM
        if ((addr >= kROM_Base) && (addr < kROM_Top))
        {
          switch (Size)
          {
            case Size_Byte : return (UBYTE) ROM_GetB ((addr - kROM_Base) % kROM_Size); break;
            case Size_Word : return (UWORD) ROM_GetW ((addr - kROM_Base) % kROM_Size); break;
            case Size_Long : return (ULONG) ROM_GetL ((addr - kROM_Base) % kROM_Size); break;
          }
        }
        if ((addr >= kRAM_Overlay_Base) && (addr < kRAM_Overlay_Top))
        {
          switch (Size)
          {
            case Size_Byte : return (UBYTE) RAM_GetB ((addr - kRAM_Overlay_Base) % kRAM_Size); break;
            case Size_Word : return (UWORD) RAM_GetW ((addr - kRAM_Overlay_Base) % kRAM_Size); break;
            case Size_Long : return (ULONG) RAM_GetL ((addr - kRAM_Overlay_Base) % kRAM_Size); break;
          }
        }
        if ((addr >= kSCSI_Block_Base) && (addr < kSCSI_Block_Top))
        {
          switch (Size)
          {
            case Size_Byte : return (UBYTE) SCSI_GetB (addr - kSCSI_Block_Base); break;
            case Size_Word : return (UWORD) SCSI_GetW (addr - kSCSI_Block_Base); break;
            case Size_Long : return (ULONG) SCSI_GetL (addr - kSCSI_Block_Base); break;
          }
        }
		#ifdef _Memory_Debug
        switch (Size)
        {
          case Size_Byte : printf("Byte Access at %8lx\n", addr); break;
          case Size_Word : printf("Word Access at %8lx\n", addr); break;
          case Size_Long : printf("Long Access at %8lx\n", addr); break;
        }
		#endif
        return 0;
        break;
      case 2 : // SCC
        if ((addr >= kSCCRd_Block_Base) && (addr < kSCCRd_Block_Top)) // Read Only Address
        {
          switch (Size)
          {
            case Size_Byte : return (UBYTE) SCC_GetB ((addr - kSCCRdBase) % kSCC_Size); break;
            case Size_Word : return (UWORD) SCC_GetW ((addr - kSCCRdBase) % kSCC_Size); break;
            case Size_Long : return (ULONG) SCC_GetL ((addr - kSCCRdBase) % kSCC_Size); break;
          }
        }
		#ifdef _Memory_Debug
        switch (Size)
        {
          case Size_Byte : printf("Byte Access at %8lx\n", addr); break;
          case Size_Word : printf("Word Access at %8lx\n", addr); break;
          case Size_Long : printf("Long Access at %8lx\n", addr); break;
        }
		#endif
		return 0;
        break;
      case 3 : // IWM and VIA
        if ((addr >= kVIA_Block_Base) && (addr < kVIA_Block_Top))
        {
          switch (Size)
          {
            case Size_Byte : return (UBYTE) VIA_GetB ((addr - kVIA_Base) % kVIA_Size); break;
            case Size_Word : return (UWORD) VIA_GetW ((addr - kVIA_Base) % kVIA_Size); break;
            case Size_Long : return (ULONG) VIA_GetL ((addr - kVIA_Base) % kVIA_Size); break;
          }
        }
        if ((addr >= kIWM_Block_Base) && (addr < kIWM_Block_Top))
        {
          switch (Size)
          {
            case Size_Byte : return (UBYTE) IWM_GetB ((addr - kIWM_Base) % kIWM_Size); break;
            case Size_Word : return (UWORD) IWM_GetW ((addr - kIWM_Base) % kIWM_Size); break;
            case Size_Long : return (ULONG) IWM_GetL ((addr - kIWM_Base) % kIWM_Size); break;
          }
        }
        if ((addr >= kAutoVector_Base) && (addr < kAutoVector_Top))
        { 
          Exception(regs.intmask+24, 0);
          return 0;
        }
		#ifdef _Memory_Debug
        switch (Size)
        {
          case Size_Byte : printf("Byte Access at %8lx\n", addr); break;
          case Size_Word : printf("Word Access at %8lx\n", addr); break;
          case Size_Long : printf("Long Access at %8lx\n", addr); break;
        }
		#endif
        return 0;
        break;
    }
  }
  else
  {
    switch (addr >> 22)
    {
      case 0 : // RAM
        if (addr < kRAM_Top)
        {
          switch (Size)
          {
            case Size_Byte : return (UBYTE) RAM_GetB (addr % kRAM_Size); break;
            case Size_Word : return (UWORD) RAM_GetW (addr % kRAM_Size); break;
            case Size_Long : return (ULONG) RAM_GetL (addr % kRAM_Size); break;
          }
        }
		#ifdef _Memory_Debug
        switch (Size)
        {
          case Size_Byte : printf("Byte Access at %8lx\n", addr); break;
          case Size_Word : printf("Word Access at %8lx\n", addr); break;
          case Size_Long : printf("Long Access at %8lx\n", addr); break;
        }
		#endif
        return 0;
        break;
      case 1 : // ROM and SCSI
        if ((addr >= kROM_Base) && (addr < kROM_Top))
        {
          switch (Size)
          {
            case Size_Byte : return (UBYTE) ROM_GetB ((addr - kROM_Base) % kROM_Size); break;
            case Size_Word : return (UWORD) ROM_GetW ((addr - kROM_Base) % kROM_Size); break;
            case Size_Long : return (ULONG) ROM_GetL ((addr - kROM_Base) % kROM_Size); break;
          }
        }
        if ((addr >= kSCSI_Block_Base) && (addr < kSCSI_Block_Top))
        {
          switch (Size)
          {
            case Size_Byte : return (UBYTE) SCSI_GetB (addr - kSCSI_Block_Base); break;
            case Size_Word : return (UWORD) SCSI_GetW (addr - kSCSI_Block_Base); break;
            case Size_Long : return (ULONG) SCSI_GetL (addr - kSCSI_Block_Base); break;
          }
        }
		#ifdef _Memory_Debug
        switch (Size)
        {
          case Size_Byte : printf("Byte Access at %8lx\n", addr); break;
          case Size_Word : printf("Word Access at %8lx\n", addr); break;
          case Size_Long : printf("Long Access at %8lx\n", addr); break;
        }
		#endif
        return 0;
        break;
      case 2 : // SCC
        if ((addr >= kSCCRd_Block_Base) && (addr < kSCCRd_Block_Top)) // Read Only Address
        {
          switch (Size)
          {
            case Size_Byte : return (UBYTE) SCC_GetB ((addr - kSCCRdBase) % kSCC_Size); break;
            case Size_Word : return (UWORD) SCC_GetW ((addr - kSCCRdBase) % kSCC_Size); break;
            case Size_Long : return (ULONG) SCC_GetL ((addr - kSCCRdBase) % kSCC_Size); break;
          }
        }
		#ifdef _Memory_Debug
        switch (Size)
        {
          case Size_Byte : printf("Byte Access at %8lx\n", addr); break;
          case Size_Word : printf("Word Access at %8lx\n", addr); break;
          case Size_Long : printf("Long Access at %8lx\n", addr); break;
        }
		#endif
		return 0;
        break;
      case 3 : // IWM and VIA
        if ((addr >= kVIA_Block_Base) && (addr < kVIA_Block_Top))
        {
          switch (Size)
          {
            case Size_Byte : return (UBYTE) VIA_GetB ((addr - kVIA_Base) % kVIA_Size); break;
            case Size_Word : return (UWORD) VIA_GetW ((addr - kVIA_Base) % kVIA_Size); break;
            case Size_Long : return (ULONG) VIA_GetL ((addr - kVIA_Base) % kVIA_Size); break;
          }
        }
        if ((addr >= kIWM_Block_Base) && (addr < kIWM_Block_Top))
        {
          switch (Size)
          {
            case Size_Byte : return (UBYTE) IWM_GetB ((addr - kIWM_Base) % kIWM_Size); break;
            case Size_Word : return (UWORD) IWM_GetW ((addr - kIWM_Base) % kIWM_Size); break;
            case Size_Long : return (ULONG) IWM_GetL ((addr - kIWM_Base) % kIWM_Size); break;
          }
        }
        if ((addr >= kAutoVector_Base) && (addr < kAutoVector_Top))
        {
          Exception(regs.intmask+24, 0);
/* Added 0 -Michael */
          return 0;
        }
		#ifdef _Memory_Debug
        switch (Size)
        {
          case Size_Byte : printf("Byte Access at %8lx\n", addr); break;
          case Size_Word : printf("Word Access at %8lx\n", addr); break;
          case Size_Long : printf("Long Access at %8lx\n", addr); break;
        }
		#endif
        return 0;
        break;
    }
  }

/* Tried to fix this too, return 0 -Michael */

  return 0;

}

void  Put_MM (CPTR addr, Size_Ty Size, ULONG in)
{

  if (vOverlay)
  {
    switch (addr >> 22)
    {
      case 0 : // ROM
        if (addr < kROM_Overlay_Top)
        {
          switch (Size)
          {
            case Size_Byte : ROM_PutB (addr % kROM_Size, (UBYTE) in); break;
            case Size_Word : ROM_PutW (addr % kROM_Size, (UWORD) in); break;
            case Size_Long : ROM_PutL (addr % kROM_Size, (ULONG) in); break;
          }
          return;
        }
		#ifdef _Memory_Debug
        switch (Size)
        {
          case Size_Byte : printf("Byte Access at %8lx, %2lx\n", addr, (UBYTE) in); break;
          case Size_Word : printf("Word Access at %8lx, %4lx\n", addr, (UWORD) in); break;
          case Size_Long : printf("Long Access at %8lx, %8lx\n", addr, (ULONG) in); break;
        }
		#endif
        break;
      case 1 : // ROM, SCSI and RAM
        if ((addr >= kROM_Base) && (addr < kROM_Top))
        {
          switch (Size)
          {
            case Size_Byte : ROM_PutB ((addr - kROM_Base) % kROM_Size, (UBYTE) in); break;
            case Size_Word : ROM_PutW ((addr - kROM_Base) % kROM_Size, (UWORD) in); break;
            case Size_Long : ROM_PutL ((addr - kROM_Base) % kROM_Size, (ULONG) in); break;
          }
          return;
        }
        if ((addr >= kRAM_Overlay_Base) && (addr < kRAM_Overlay_Top))
        {
          switch (Size)
          {
            case Size_Byte : RAM_PutB ((addr - kRAM_Overlay_Base) % kRAM_Size, (UBYTE) in); break;
            case Size_Word : RAM_PutW ((addr - kRAM_Overlay_Base) % kRAM_Size, (UWORD) in); break;
            case Size_Long : RAM_PutL ((addr - kRAM_Overlay_Base) % kRAM_Size, (ULONG) in); break;
          }
          return;
        }
        if ((addr >= kSCSI_Block_Base) && (addr < kSCSI_Block_Top))
        {
          switch (Size)
          {
            case Size_Byte : SCSI_PutB (addr - kSCSI_Block_Base, (UBYTE) in); break;
            case Size_Word : SCSI_PutW (addr - kSCSI_Block_Base, (UWORD) in); break;
            case Size_Long : SCSI_PutL (addr - kSCSI_Block_Base, (ULONG) in); break;
          }
          return;
        }
		#ifdef _Memory_Debug
        switch (Size)
        {
          case Size_Byte : printf("Byte Access at %8lx, %2lx\n", addr, (UBYTE) in); break;
          case Size_Word : printf("Word Access at %8lx, %4lx\n", addr, (UWORD) in); break;
          case Size_Long : printf("Long Access at %8lx, %8lx\n", addr, (ULONG) in); break;
        }
		#endif
        break;
      case 2 : // SCC
        if ((addr >= kSCCWr_Block_Base) && (addr < kSCCWr_Block_Top)) // Write Only Address
        {
          switch (Size)
          {
            case Size_Byte : SCC_PutB ((addr - kSCCWrBase) % kSCC_Size, (UBYTE) in); break;
            case Size_Word : SCC_PutW ((addr - kSCCWrBase) % kSCC_Size, (UWORD) in); break;
            case Size_Long : SCC_PutL ((addr - kSCCWrBase) % kSCC_Size, (ULONG) in); break;
          }
          return;
        }
		#ifdef _Memory_Debug
        switch (Size)
        {
          case Size_Byte : printf("Byte Access at %8lx, %2lx\n", addr, (UBYTE) in); break;
          case Size_Word : printf("Word Access at %8lx, %4lx\n", addr, (UWORD) in); break;
          case Size_Long : printf("Long Access at %8lx, %8lx\n", addr, (ULONG) in); break;
        }
		#endif
        break;
      case 3 : // IWM and VIA
        if ((addr >= kVIA_Block_Base) && (addr < kVIA_Block_Top))
        {
          switch (Size)
          {
            case Size_Byte : VIA_PutB ((addr - kVIA_Base) % kVIA_Size, (UBYTE) in); break;
            case Size_Word : VIA_PutW ((addr - kVIA_Base) % kVIA_Size, (UWORD) in); break;
            case Size_Long : VIA_PutL ((addr - kVIA_Base) % kVIA_Size, (ULONG) in); break;
          }
          return;
        }
        if ((addr >= kIWM_Block_Base) && (addr < kIWM_Block_Top))
        {
          switch (Size)
          {
            case Size_Byte : IWM_PutB ((addr - kIWM_Base) % kIWM_Size, (UBYTE) in); break;
            case Size_Word : IWM_PutW ((addr - kIWM_Base) % kIWM_Size, (UWORD) in); break;
            case Size_Long : IWM_PutL ((addr - kIWM_Base) % kIWM_Size, (ULONG) in); break;
          }
          return;
        }
        if ((addr >= kAutoVector_Base) && (addr < kAutoVector_Top))
        {
          Exception(regs.intmask+24, 0);
          return;
        }
		#ifdef _Memory_Debug
        switch (Size)
        {
          case Size_Byte : printf("Byte Access at %8lx, %2lx\n", addr, (UBYTE) in); break;
          case Size_Word : printf("Word Access at %8lx, %4lx\n", addr, (UWORD) in); break;
          case Size_Long : printf("Long Access at %8lx, %8lx\n", addr, (ULONG) in); break;
        }
		#endif
        break;
    }
  }
  else
  {
    switch (addr >> 22)
    {
      case 0 : // RAM
        if (addr < kRAM_Top)
        {
          switch (Size)
          {
            case Size_Byte : RAM_PutB (addr % kRAM_Size, (UBYTE) in); break;
            case Size_Word : RAM_PutW (addr % kRAM_Size, (UWORD) in); break;
            case Size_Long : RAM_PutL (addr % kRAM_Size, (ULONG) in); break;
          }
          return;
        }
		#ifdef _Memory_Debug
        switch (Size)
        {
          case Size_Byte : printf("Byte Access at %8lx, %2lx\n", addr, (UBYTE) in); break;
          case Size_Word : printf("Word Access at %8lx, %4lx\n", addr, (UWORD) in); break;
          case Size_Long : printf("Long Access at %8lx, %8lx\n", addr, (ULONG) in); break;
        }
		#endif
        break;
      case 1 : // ROM and SCSI
        if ((addr >= kROM_Base) && (addr < kROM_Top))
        {
          switch (Size)
          {
            case Size_Byte : ROM_PutB ((addr - kROM_Base) % kROM_Size, (UBYTE) in); break;
            case Size_Word : ROM_PutW ((addr - kROM_Base) % kROM_Size, (UWORD) in); break;
            case Size_Long : ROM_PutL ((addr - kROM_Base) % kROM_Size, (ULONG) in); break;
          }
          return;
        }
        if ((addr >= kSCSI_Block_Base) && (addr < kSCSI_Block_Top))
        {
          switch (Size)
          {
            case Size_Byte : SCSI_PutB (addr - kSCSI_Block_Base, (UBYTE) in); break;
            case Size_Word : SCSI_PutW (addr - kSCSI_Block_Base, (UWORD) in); break;
            case Size_Long : SCSI_PutL (addr - kSCSI_Block_Base, (ULONG) in); break;
          }
          return;
        }
		#ifdef _Memory_Debug
        switch (Size)
        {
          case Size_Byte : printf("Byte Access at %8lx, %2lx\n", addr, (UBYTE) in); break;
          case Size_Word : printf("Word Access at %8lx, %4lx\n", addr, (UWORD) in); break;
          case Size_Long : printf("Long Access at %8lx, %8lx\n", addr, (ULONG) in); break;
        }
		#endif
        break;
      case 2 : // SCC
        if ((addr >= kSCCWr_Block_Base) && (addr < kSCCWr_Block_Top))
        {
          switch (Size)
          {
            case Size_Byte : SCC_PutB ((addr - kSCCWrBase) % kSCC_Size, (UBYTE) in); break;
            case Size_Word : SCC_PutW ((addr - kSCCWrBase) % kSCC_Size, (UWORD) in); break;
            case Size_Long : SCC_PutL ((addr - kSCCWrBase) % kSCC_Size, (ULONG) in); break;
          }
          return;
        }
		#ifdef _Memory_Debug
        switch (Size)
        {
          case Size_Byte : printf("Byte Access at %8lx, %2lx\n", addr, (UBYTE) in); break;
          case Size_Word : printf("Word Access at %8lx, %4lx\n", addr, (UWORD) in); break;
          case Size_Long : printf("Long Access at %8lx, %8lx\n", addr, (ULONG) in); break;
        }
		#endif
        break;
      case 3 : // IWM and VIA
	    if ((addr >= kVIA_Block_Base) && (addr < kVIA_Block_Top))
        {
          switch (Size)
          {
            case Size_Byte : VIA_PutB ((addr - kVIA_Base) % kVIA_Size, (UBYTE) in); break;
            case Size_Word : VIA_PutW ((addr - kVIA_Base) % kVIA_Size, (UWORD) in); break;
            case Size_Long : VIA_PutL ((addr - kVIA_Base) % kVIA_Size, (ULONG) in); break;
          }
          return;
        }
        if ((addr >= kIWM_Block_Base) && (addr < kIWM_Block_Top))
        {
          switch (Size)
          {
            case Size_Byte : IWM_PutB ((addr - kIWM_Base) % kIWM_Size, (UBYTE) in); break;
            case Size_Word : IWM_PutW ((addr - kIWM_Base) % kIWM_Size, (UWORD) in); break;
            case Size_Long : IWM_PutL ((addr - kIWM_Base) % kIWM_Size, (ULONG) in); break;
          }
          return;
        }
        if ((addr >= kAutoVector_Base) && (addr < kAutoVector_Top))
        {
          Exception(regs.intmask+24, 0);
          return;
        }
		#ifdef _Memory_Debug
        switch (Size)
        {
          case Size_Byte : printf("Byte Access at %8lx, %2lx\n", addr, (UBYTE) in); break;
          case Size_Word : printf("Word Access at %8lx, %4lx\n", addr, (UWORD) in); break;
          case Size_Long : printf("Long Access at %8lx, %8lx\n", addr, (ULONG) in); break;
        }
		#endif
        break;
    }
  }
}

// UAE 0.6 Emulation Direct Memory Access Function

/* was UWORD */

UWORD *device_old_interface_get_real_address (CPTR addr)
{
  CPTR taddr = addr & kMemory_Extent;
  
  if (vOverlay)
  {
    switch (taddr >> 22)
    {
      case 0 : // ROM
        if (taddr < kROM_Overlay_Top)
        {
          return ROM_Xlate (taddr % kROM_Size);
        }
		return NULL;
        break;
      case 1 : // ROM, SCSI and RAM
        if ((taddr >= kROM_Base) && (taddr < kROM_Top))
        {
          return ROM_Xlate ((taddr - kROM_Base) % kROM_Size);
        }
        if ((taddr >= kRAM_Overlay_Base) && (taddr < kRAM_Overlay_Top))
        {
          return RAM_Xlate ((taddr - kRAM_Overlay_Base) % kRAM_Size);
        }
        return NULL;
        break;
      case 2 : // SCC
        return NULL;
        break;
      case 3 : // IWM and VIA
        return NULL;
        break;
    }
  }
  else
  {
    switch (taddr >> 22)
    {
      case 0 : // RAM
        if (taddr < kRAM_Top)
        {
          return RAM_Xlate (taddr % kRAM_Size);
        }
		return NULL;
        break;
      case 1 : // ROM and SCSI
        if ((taddr >= kROM_Base) && (taddr < kROM_Top))
        {
          return ROM_Xlate ((taddr - kROM_Base) % kROM_Size);
        }
        return NULL;
        break;
      case 2 : // SCC
        return NULL;
        break;
      case 3 : // IWM and VIA
        return NULL;
        break;
    }
  }
/* tried to fix this too... -Michael */

  return 0;

}

// UAE 0.6 68K Emulator Memory Access Functions

//ULONG get_long(CPTR addr)
ULONG device_old_interface_get_long(CPTR addr)
{
  return (ULONG) Get_MM(addr & kMemory_Extent,Size_Long);
}

UWORD device_old_interface_get_word(CPTR addr)
{
  return (UWORD) Get_MM(addr & kMemory_Extent,Size_Word);
}

UBYTE device_old_interface_get_byte(CPTR addr)
{
  return (UBYTE) Get_MM(addr & kMemory_Extent,Size_Byte);
}

void device_old_interface_put_long(CPTR addr, ULONG in)
{
  Put_MM(addr & kMemory_Extent, Size_Long, (ULONG) in);
}

void device_old_interface_put_word(CPTR addr, UWORD in)
{
  Put_MM(addr & kMemory_Extent, Size_Word, (ULONG) in);
}

void device_old_interface_put_byte(CPTR addr, UBYTE in)
{
  Put_MM(addr & kMemory_Extent, Size_Byte, (ULONG) in);
}

// VIA Interface Functions

UBYTE VIA_GORA4 (void) // Overlay/Normal Memory Mapping
{
  #ifdef _VIA_Interface_Debug
  printf("VIA ORA4 attempts to be an input\n");
  #endif
  return 0;
}

void  VIA_PORA4 (UBYTE Data)
{
	vOverlay = Data;
	memory_management_overlay(Data);
}

void memory_init_uae_069(void)
{
	memory_init();
}

void memory_init(void)
{
	int i;
	buserr = 0;
	RAMmem_mask		= kRAM_Size - 1;
	ROMmem_mask		= kROM_Size - 1;
	Overlay_RAMmem_mask	= Overlay_RAMmem_size - 1;
	Overlay_ROMmem_mask	= Overlay_ROMmem_size - 1;

	RAMmemory = Overlay_RAMmemory = (UBYTE *) RAM;
	ROMmemory = Overlay_ROMmemory = (UBYTE *) ROM;

	mem_banks = Non_overlay_mem_banks;

	for(i = 0; i < 65536; i++)
		put_mem_bank (i<<16, &dummy_bank);

// First 4 megabytes of RAM
	map_banks(&RAMmem_bank   , 0x0000, 0x0040 - 0x0000);

	// Non-overlay	ROM layout

	map_banks(&ROMmem_bank   , 0x0040, 0x0050 - 0x0040); // 128k Macintosh Plus ROM
	map_banks(&devicemem_bank, 0x0058, 0x0060 - 0x0058); // SCSI	device
	map_banks(&devicemem_bank, 0x0080, 0x00a0 - 0x0080); // SCC	device read
	map_banks(&devicemem_bank, 0x00e8, 0x00f0 - 0x00e8); // VIA	device
	map_banks(&devicemem_bank, 0x00c0, 0x00e0 - 0x00c0); // IWM	device
	map_banks(&devicemem_bank, 0x00ff, 0x0100 - 0x00ff); // AutoVector	device
	map_banks(&devicemem_bank, 0x00a0, 0x00c0 - 0x00a0); // SCC  device write

	// Clone the currently active memory bank

	for(i = 0; i < 65536; i++)
	{
		Overlay_mem_banks[i] = Non_overlay_mem_banks[i];
	}

	mem_banks = Overlay_mem_banks;

	map_banks(&Overlay_RAMmem_bank, 0x0060, 0x0070 - 0x0060);

	map_banks(&dummy_bank	      , 0x0000, 0x0040 - 0x0000);
	map_banks(&Overlay_ROMmem_bank, 0x0000, 0x0010 - 0x0000);

	mem_banks = Overlay_mem_banks;

	ROMmem_write_enable_set(off);
}

void memory_management_overlay(int test)
{
	if (test)
	{
		mem_banks = Overlay_mem_banks;
	}
	else
	{
		mem_banks = Non_overlay_mem_banks;
	}
}

void ROMmem_write_enable_set(unsigned int goo)
{
	if (goo)
	{
		ROMmem_write_enable = 1;
	}
	else
	{
		ROMmem_write_enable = 0;
	}
}

char *address_space;

void set_uae_vmac_address_space()
{
        //casting from a UWORD...
        address_space= (char *) RAM;
}

