// vMac 0.x Source
// By Philip Cummins (PC)
// Copyright PC 1997
   
// PC - ROM Function Files

#include "sysconfig.h"
#include "sysdeps.h"
#include "config.h"
#include "ROM.h"
#include "prefs.h"

#include "memory.h"

#ifdef _ROM_Debug
#include <stdio.h>
#endif

#include "Unix/universal.h"

UWORD private_ROM [kROM_Size/2];
UWORD *ROM = private_ROM;

int  ROM_Init (void)
{
  FILE *ROM_File;
  int  File_Size;
  int i;

  /* Gemulator ROM board support */
  if(thePrefs.CurrentROMPath==255)
 {
      printf("vMac: Using hardware ROM (Gemulator ROM board)\n");
 
      File_Size=Gemulator_LoadROM(ROM, thePrefs.GemulatorBase, thePrefs.GemulatorSocket, 2); //2 for MacPlus ROM set

      if (File_Size < kROM_Size)
      {
        printf("vMac: Wrong Gemulator ROM board setting(s) in .vmacrc\n");
        return 1;
      }
 }
  else
 {
  printf("vMac: Using ROM image file\n");
 
  ROM_File = fopen(thePrefs.ROMPath[thePrefs.CurrentROMPath], "rb");
  if (ROM_File == NULL) 
    ROM_File = fopen("vMac.ROM","rb");
  if (ROM_File == NULL)
  {    
    printf("vMac: The vMac.ROM could not be found.\n");
    return 1;
  }
  printf("vMac: Loading vMac.ROM...\n");
  File_Size = fread(ROM, kROM_Size, 1, ROM_File);
#ifndef MKLINUX_GLIBC_BUG
  if (File_Size != 1)
  {
    printf("vMac: The vMac.ROM is not 128 KB.\n");
    fclose(ROM_File);
    return 1;
  }
#endif
  fclose(ROM_File);
}

  if ((do_get_mem_word(&ROM[0]) == 0x4D1E)
   && (do_get_mem_word(&ROM[1]) == 0xEEE1))
  {
    printf("vMac: vMac.ROM loaded successfully, Mac Plus ROM v 1, 'Lonely Hearts'.\n");
    return 0;
  }
  if ((do_get_mem_word(&ROM[0]) == 0x4D1E)
   && (do_get_mem_word(&ROM[1]) == 0xEAE1))
  {
    printf("vMac: vMac.ROM loaded successfully, Mac Plus ROM v 2, 'Lonely Heifers'.\n");
    return 0;
  }
  if ((do_get_mem_word(&ROM[0]) == 0x4D1F)
   && (do_get_mem_word(&ROM[1]) == 0x8172))
  {
    printf("vMac: vMac.ROM loaded successfully, Mac Plus ROM v 3, 'Loud Harmonicas'.\n");

#ifdef ROM_CHANGE_TEST
    printf("value1: %x\n", do_get_mem_word(&ROM[0x11e]));
    printf("value2: %x\n", do_get_mem_word(&ROM[0x120]));

    do_put_mem_word(&ROM[0x11e], 0x4e71);
    do_put_mem_word(&ROM[0x120], 0x4e71);

    printf("value1-changed: %x\n", do_get_mem_word(&ROM[0x11e]));
    printf("value2-changed: %x\n", do_get_mem_word(&ROM[0x120]));
#endif

    return 0;
  }
  printf("vMac: vMac.ROM loaded successfully, but we don't support this ROM yet.\n");
  printf("vMac: ROM Checksum: %X %X\n",do_get_mem_word(&ROM[0]),do_get_mem_word(&ROM[1]));
  return 1;
}

void  ROM_Reset (void)
{
  extern UBYTE Sony_Installed;
  
  if (Sony_Installed == 1)
    (void) ROM_Init();
}

UWORD *ROM_Xlate (CPTR addr)
{
  return ROM + (addr >> 1);
}

UBYTE ROM_GetB (CPTR addr)
{
  return ROM [addr >> 1] >> (addr & 1 ? 0 : 8);
}

UWORD ROM_GetW (CPTR addr)
{
  return ROM [addr >> 1];
}

ULONG ROM_GetL (CPTR addr)
{
  return ((ULONG) ROM [addr >> 1] << 16) | ROM [(addr >> 1) + 1];
}

void  ROM_PutB (CPTR addr, UBYTE in)
{
  return;
}

void  ROM_PutW (CPTR addr, UWORD in)
{
  return;
}

void  ROM_PutL (CPTR addr, ULONG in)
{
  return;
}

