// PC - Sony Driver Functions

#include "sysconfig.h"
#include "sysdeps.h"
#include "config.h"
#include "options.h"

#include "memory.h"
#include "Memory.h"
#include "custom.h"

#include "readcpu.h"
#include "newcpu.h"

#include "Native.h"
#include "Sony.h"
#include "prefs.h"

#include "Unix/universal.h"
#include "reentrant_newcpu.h"

#include <ctype.h>
#include "debug.h"

#ifdef _Sony_Debug
#include <stdio.h>
#endif

#define _Sony_Writable

UBYTE Sony_Installed = 0;
UBYTE Sony_Opened = 0;
CPTR Sony_Table[5];
CPTR ParamBlk = 0;
CPTR DeviceCtl = 0;

FILE *Drive1 = NULL; // File link to Drive 1
FILE *Drive2 = NULL; // File link to Drive 2
FILE *Drive3 = NULL; // File link to Drive 3

int Sony_Init (void)
{
  Sony_Installed = 0;
  Sony_Opened = 0;
  if (Drive1 != NULL)
    fclose(Drive1);
  if (Drive2 != NULL)
    fclose(Drive2);
  if (Drive3 != NULL)
    fclose(Drive3);
  return 0;
}

void Sony_Reset (void)
{
  (void) Sony_Init();
}

// Check_Sony_Patch
// Checks to see if the Sony driver is being read, then patches it up with Sony_Install
int Check_Sony_Patch (void)
{  
CPTR NamePtr = 0;
UWORD *SonyPtr = NULL;

/*
	if ( get_word(m68k_getpc())
		== (UWORD) 0xA937)
	{
		printf("DrawMenuBar called.\n");
	}
*/ /* Michael- Just playing with A Line traps. */

	if (Sony_Installed == 0)
	{

// Check if it's the A line trap "_GetNamedResource"
		if ( get_word(m68k_getpc())
				== (UWORD) 0xA9A1)
		{
			NamePtr = get_long(m68k_areg(regs, 7));

			if(	   (get_long(NamePtr)	== 0x052E536F)
				&& (get_word(NamePtr+4)	== 0x6E79)
				&& (get_long(m68k_areg(regs, 7) + 4) )
				== 0x44525652)
			{ 

				Run_ATrap(); // Call GetNamedResource

				Sony_Install();

				return 1;
			}
		}
		else
		{
			return 0;
		}
	}

	return 0; 
}

// Check_Sony_Illegal
// Checks to see if the ILLEGAL was a call to our Sony driver
int Check_Sony_Illegal (void)
{


  if (Sony_Table[kOpen] == m68k_getpc())
  {
    Sony_Open();
    return 1;
  }
  if (Sony_Table[kPrime] == m68k_getpc())
  {
    Sony_Prime();
    return 1;
  }
  if (Sony_Table[kControl] == m68k_getpc())
  {
    Sony_Control();
    return 1;
  }
  if (Sony_Table[kStatus] == m68k_getpc())
  {
    Sony_Status();
    return 1;
  }
  if (Sony_Table[kClose] == m68k_getpc())
  {
    Sony_Close();
    return 1;
  }
  return 0;
}

// Sony_Install
// This installs a dummy ILLEGAL+RTS header over the Sony driver
// It allows us to trap ILLEGAL to our own routines
void Sony_Install (void)
{
  CPTR VectorPos = 0, SonyPtr = 0;
  int Counter = 0;

  VectorPos = SonyPtr = get_long(get_long(m68k_areg(regs, 7))) & 0x00FFFFFF;

ROMmem_write_enable_set(on);

	SonyPtr += sizeof(UWORD) * 4; // Skip Driver Header (4 Words)

	put_word(SonyPtr, 0x0018); SonyPtr += sizeof(UWORD); // Adjust Open Offset
	put_word(SonyPtr, 0x001C); SonyPtr += sizeof(UWORD); // Adjust Prime Offset
	put_word(SonyPtr, 0x0020); SonyPtr += sizeof(UWORD); // Adjust Control Offset
	put_word(SonyPtr, 0x0024); SonyPtr += sizeof(UWORD); // Adjust Status Offset
	put_word(SonyPtr, 0x0028); SonyPtr += sizeof(UWORD); // Adjust Close Offset

	SonyPtr += sizeof(UWORD) * 3; // Skip Driver Name (3 Words)

	VectorPos += 0x0018; // Update Vector Position

	for (Counter = 0; Counter <= 4; Counter++)
	{
// Install ILLEGAL instruction
		put_word(SonyPtr, 0x4AFC); SonyPtr += sizeof(UWORD);
// Install RTS instruction
		put_word(SonyPtr, 0x4E75); SonyPtr += sizeof(UWORD);

		Sony_Table[Counter] = VectorPos;
		VectorPos += 4;
	}

ROMmem_write_enable_set(off);

	Sony_Installed = 1;
}

// Sony_Return
// Returns from a Sony driver call properly
// Sets D0 to be the Error code, or leaves it alone if SaveD0 = 1
// If Mode = 1, this is an immediate call for sure, use RTS to exit
// Otherwise check it ourselves and RTS or jIODone
void Sony_Return (int iErr, int SaveD0, int Mode)
{
  put_word(ParamBlk + kioResult, iErr); // Update ioResult
  if (SaveD0 == 0) { // Don't Save D0
    if ((iErr & 0x8000) == 0x8000)
      m68k_dreg(regs, 0) = 0xFFFF0000 | iErr; // Mask D0 properly
    else
      m68k_dreg(regs, 0) = 0x00000000 | iErr; // Clear D0 properly
  }
  // Synchronous, Asynchronous or Immediate
  if (((get_word(ParamBlk + kioTrap) & 0x0200) == 0x0200) || (Mode == 1)) 
    m68k_setpc(m68k_getpc() + 2); // Immediate Return (RTS)
  else
    m68k_setpc(get_long(0x08FC)); // Jump to JIODone (always done)
}

void Sony_Insert (UWORD Drive_No, char *filename)
{
  CPTR SonyVars = get_long(0x0134);
  ULONG tA0 = m68k_areg(regs, 0);
  ULONG tD0 = m68k_dreg(regs, 0);
  
  if (Sony_Opened == 0)
    return; // Sony driver isn't open yet, ignore
  
  if (Drive_No == 0x0001)
  {
    #ifdef _Sony_Writable
#ifdef NeXT
    if( Drive1 != NULL ) {
        fflush(Drive1);
        fclose(Drive1);
    }
#else
    fflush(Drive1);
#endif
    Drive1 = fopen(filename, "rb+");
    if (Drive1 == NULL)
    {
      return; // Error, no such disk file
    }
    put_byte(SonyVars + kDrive1 + kWriteProt, 0x00);   // Drive 1 Unlocked
    #else
#ifdef NeXT
    if( Drive1 != NULL ) {
        fflush(Drive1);  
        fclose(Drive1);
    }
#else
    fflush(Drive1);
#endif
    Drive1 = fopen(filename, "rb");
    if (Drive1 == NULL)
    {
      return; // Error, no such disk file
    }
    put_byte(SonyVars + kDrive1 + kWriteProt, 0xFF);   // Drive 1 Locked
    #endif
    put_byte(SonyVars + kDrive1 + kDiskInPlace, 0x01); // Drive 1 Disk Inserted
    put_byte(SonyVars + kDrive1 + kTwoSideFmt, 0xFF);  // Drive 1 Double Format
  }
  if (Drive_No == 0x0002)
  {
    #ifdef _Sony_Writable
#ifdef NeXT
    if( Drive2 != NULL ) {
        fflush(Drive2);
        fclose(Drive2);
    }
#else
    fflush(Drive2);
#endif
    Drive2 = fopen(filename, "rb+");
    if (Drive2 == NULL)
    {
      return; // Error, no such disk file
    }
    put_byte(SonyVars + kDrive2 + kWriteProt, 0x00);   // Drive 2 Unlocked
    #else
#ifdef NeXT
    if( Drive2 != NULL ) {
        fflush(Drive2);  
        fclose(Drive2);
    }
#else
    fflush(Drive2);
#endif
    Drive2 = fopen(filename, "rb");
    if (Drive2 == NULL)
    {
      return; // Error, no such disk file
    }
    put_byte(SonyVars + kDrive2 + kWriteProt, 0xFF);   // Drive 2 Locked
    #endif
    put_byte(SonyVars + kDrive2 + kDiskInPlace, 0x01); // Drive 2 Disk Inserted
    put_byte(SonyVars + kDrive2 + kTwoSideFmt, 0xFF);  // Drive 2 Double Format
  }
  if (Drive_No == 0x0003)
  {
    #ifdef _Sony_Writable
#ifdef NeXT
    if( Drive3 != NULL ) {
        fflush(Drive3);
        fclose(Drive3);
    }
#else
    fflush(Drive3);
#endif
    Drive3 = fopen(filename, "rb+");
    if (Drive3 == NULL)
    {
      return; // Error, no such disk file
    }
    put_byte(SonyVars + kDrive3 + kWriteProt, 0x00);   // Drive 2 Unlocked
    #else
#ifdef NeXT
    if( Drive3 != NULL ) {
        fflush(Drive3);  
        fclose(Drive3);
    }
#else
    fflush(Drive3);
#endif
    Drive3 = fopen(filename, "rb");
    if (Drive3 == NULL)
    {
      return; // Error, no such disk file
    }
    put_byte(SonyVars + kDrive3 + kWriteProt, 0xFF);   // Drive 2 Locked
    #endif
    put_byte(SonyVars + kDrive3 + kDiskInPlace, 0x01); // Drive 2 Disk Inserted
    put_byte(SonyVars + kDrive3 + kTwoSideFmt, 0xFF);  // Drive 2 Double Format
  }
  Sony_Update();
}

void Sony_Eject (UWORD Drive_No)
{
  CPTR SonyVars = get_long(0x0134);
  
  if (Drive_No == 0x0001)
  {
    put_byte(SonyVars + kDrive1 + kWriteProt, 0x00);   // Drive 1 Writeable
    put_byte(SonyVars + kDrive1 + kDiskInPlace, 0x00); // Drive 1 No Disk
    put_byte(SonyVars + kDrive1 + kTwoSideFmt, 0x00);  // Drive 1 Single Format (Initially)
    fclose(Drive1);
    Drive1 = NULL;
    return;
  }
  if (Drive_No == 0x0002)
  {
    put_byte(SonyVars + kDrive2 + kWriteProt, 0x00);   // Drive 2 Writeable
    put_byte(SonyVars + kDrive2 + kDiskInPlace, 0x00); // Drive 2 No Disk
    put_byte(SonyVars + kDrive2 + kTwoSideFmt, 0x00);  // Drive 2 Single Format (Initially)
    fclose(Drive2);
    Drive2 = NULL;
    return;
  }
  if (Drive_No == 0x0003)
  {
    put_byte(SonyVars + kDrive3 + kWriteProt, 0x00);   // Drive 2 Writeable
    put_byte(SonyVars + kDrive3 + kDiskInPlace, 0x00); // Drive 2 No Disk
    put_byte(SonyVars + kDrive3 + kTwoSideFmt, 0x00);  // Drive 2 Single Format (Initially)
    fclose(Drive3);
    Drive3 = NULL;
    return;
  }
}

void Sony_Access (UWORD Drive_No)
{
  UWORD *Buffer = NULL;
  ULONG Sony_Start = 0;
  ULONG Sony_Count = 0;

  ParamBlk = m68k_areg(regs, 0);
  DeviceCtl = m68k_areg(regs, 1);
  
  Buffer = (void*)((ULONG) get_real_address(0) + (get_long(ParamBlk + kioBuffer) & kMemory_Extent));

  switch (get_word(ParamBlk + kioPosMode) & 0x0F)
  {
    case kfsAtMark :
      Sony_Start = get_long(DeviceCtl + kdCtlPosition); break;
    case kfsFromStart :
      Sony_Start = get_long(ParamBlk + kioPosOffset); break;
    case kfsFromLEOF :
      Sony_Start = 819200 - get_long(ParamBlk + kioPosOffset); break; // Gotta fix this!
    case kfsFromMark :
      Sony_Start = get_long(ParamBlk + kioPosOffset) +
                   get_long(DeviceCtl + kdCtlPosition); break;
  }  
      
  Sony_Count = get_long(ParamBlk + kioReqCount);
  
  if ( (get_word(ParamBlk + kioTrap) & 0xF0FF ) == 0xA002) // Reading Trap
  {
    if (Drive_No == 1)
    {
      Sony_LEDs(1,0,((Sony_Count >> 16) / 2));
      fseek(Drive1, Sony_Start, SEEK_SET);
      fread(Buffer, 1, Sony_Count, Drive1);
    }
  
    if (Drive_No == 2)
    {
      Sony_LEDs(2,0,((Sony_Count >> 16) / 2));
      fseek(Drive2, Sony_Start, SEEK_SET);
      fread(Buffer, 1, Sony_Count, Drive2);
    }
    if (Drive_No == 3)
    {
      Sony_LEDs(3,0,((Sony_Count >> 16) / 2));
      fseek(Drive3, Sony_Start, SEEK_SET);
      fread(Buffer, 1, Sony_Count, Drive3);
    }

  }
  
  if ( (get_word(ParamBlk + kioTrap) & 0xF0FF ) == 0xA003) // Writing
  {
    if (Drive_No == 1)
    {
      Sony_LEDs(1,1,((Sony_Count >> 16) / 2));
      fseek(Drive1, Sony_Start, SEEK_SET);
      fwrite(Buffer, 1, Sony_Count, Drive1);
    }
    if (Drive_No == 2)
    {
      Sony_LEDs(2,1,((Sony_Count >> 16) / 2));
      fseek(Drive2, Sony_Start, SEEK_SET);
      fwrite(Buffer, 1, Sony_Count, Drive2);
    }
    if (Drive_No == 3)
    {
      Sony_LEDs(3,1,((Sony_Count >> 16) / 2));
      fseek(Drive3, Sony_Start, SEEK_SET);
      fwrite(Buffer, 1, Sony_Count, Drive3);
    }
  }
  
  put_word(ParamBlk + kioResult, 0x0000);
  put_long(ParamBlk + kioActCount, Sony_Count);
  put_long(DeviceCtl + kdCtlPosition, get_long(DeviceCtl + kdCtlPosition) + Sony_Count);
  
}

int Sony_Drive_Installed (UWORD Drive_No)
{
  if ((Drive_No == 1) || (Drive_No == 2) || (Drive_No == 3))
    return 1; // Drive is installed
  else
    return 0; // Drive is not installed
}

int Sony_Disk_Inserted (UWORD Drive_No)
{
  if (Drive_No == 1) {
    if (Drive1 == NULL)
      return 0; // Disk not inserted
    else
      return 1; // Disk inserted
  }
  if (Drive_No == 2) {
    if (Drive2 == NULL)
      return 0; // Disk not inserted
    else
      return 1; // Disk inserted
  }
  if (Drive_No == 3) {
    if (Drive3 == NULL)
      return 0; // Disk not inserted
    else
      return 1; // Disk inserted
  }
// this probablly will screw things up
  return -1;
}

// This checks every VBL to see if the disk has been read
void Sony_Update (void)
{

  CPTR SonyVars = get_long(0x0134);
  ULONG tA0 = m68k_areg(regs, 0);
  ULONG tD0 = m68k_dreg(regs, 0);
  
  if (Sony_Opened == 1)
  {
    if (get_byte(SonyVars + kDrive1 + kDiskInPlace) == 0x01)
    {
      m68k_areg(regs, 0) = 0x00000007; // Disk Inserted
      m68k_dreg(regs, 0) = 0x00000001; // Drive 1

      Call_ATrap(0xA02F); // PostEvent Trap

      m68k_areg(regs, 0) = tA0;
      m68k_dreg(regs, 0) = tD0;
    }

    if (get_byte(SonyVars + kDrive2 + kDiskInPlace) == 0x01)
    {
      m68k_areg(regs, 0) = 0x00000007; // Disk Inserted
      m68k_dreg(regs, 0) = 0x00000002; // Drive 2

      Call_ATrap(0xA02F); // PostEvent Trap (bill comment ---)

      m68k_areg(regs, 0) = tA0;
      m68k_dreg(regs, 0) = tD0;
    }

    if (get_byte(SonyVars + kDrive3 + kDiskInPlace) == 0x01)
    {
      m68k_areg(regs, 0) = 0x00000007; // Disk Inserted
      m68k_dreg(regs, 0) = 0x00000003; // Drive 2

      Call_ATrap(0xA02F); // PostEvent Trap (bill comment ---)

      m68k_areg(regs, 0) = tA0;
      m68k_dreg(regs, 0) = tD0;
    }
/*
*/
  }    
}

void Sony_LEDs (UWORD Drive_No, UBYTE Mode, UWORD Time)
{
  // Break out to MacOS Code
  // Time is the number of 512 byte blocks to be transferred
  // Drive is the drive accessed
  // Mode is the mode (0 = Read, 1 = Write)
  return;
}

// Sony_Open
//
// This opens up the Sony driver, it allocates a pointer for the variables
// Sets SonyVars, then sets up the drive variables with correct values
// And then adds the drives into the drive queue
//
// Bill Huey comment ---
//
// It wasn't obvious initially, but what "Opening" the ".Sony" driver doesn't
// mean opening a file, which is pretty stupid of me. I keeped wondering
// where and how the drive structures where being set at start up, only to realize
// that it's right under my nose in the function Sony_Open.
// That's what I get for not paying attention...
//
void Sony_Open (void)
{
  CPTR tA0 = m68k_areg(regs, 0);
  CPTR tD0 = m68k_dreg(regs, 0);
  CPTR SonyVars = 0;
  
  if (Sony_Opened == 1) // Don't Open Again
  {
    Sony_Return(0x0000,0,1); // Immediate Return, No Error
    return;
  }
  
  Sony_Opened = 1;
    
  tA0 = m68k_areg(regs, 0);
  tD0 = m68k_dreg(regs, 0);
  
  m68k_areg(regs, 0) = 0x00000000;
  m68k_dreg(regs, 0) = 0x00000310; // Size of Sony Variables

// bill huey ---
//
// The HANDLE is return in the A0 register...
// The error code is return in the D0 register...
//

  Call_ATrap(0xA71E); // _NewPtr,SYS,CLEAR Trap (bill comment ---)

  put_long(0x0134,m68k_areg(regs, 0)); // Save it to SonyVars
  SonyVars = m68k_areg(regs, 0);
 
  // Add Drive 1
  put_word(SonyVars + kDrive1 + kTrack, 0x0000);     // Drive 1 at Track 0
  put_byte(SonyVars + kDrive1 + kWriteProt, 0x00);   // Drive 1 Writeable
  put_byte(SonyVars + kDrive1 + kDiskInPlace, 0x00); // Drive 1 No Disk
  put_byte(SonyVars + kDrive1 + kInstalled, 0x01);   // Drive 1 Installed
  put_byte(SonyVars + kDrive1 + kSides, 0xFF);       // Drive 1 Double Sided
  put_word(SonyVars + kDrive1 + kQDriveNo, 0x0001);  // Drive 1 is Drive 1
  put_word(SonyVars + kDrive1 + kQRefNum, 0xFFFB);   // Drive 1 uses .Sony
  put_byte(SonyVars + kDrive1 + kTwoSideFmt, 0x00);  // Drive 1 Single Format (Initially)
  put_byte(SonyVars + kDrive1 + kNewIntf, 0xFF);     // Drive 1 uses new interface
  put_word(SonyVars + kDrive1 + kDriveErrs, 0x0000); // Drive 1 has no errors
  
  m68k_areg(regs, 0) = SonyVars + kDrive1 + kQLink;
  m68k_dreg(regs, 0) = 0x0001FFFB; // Drive 1 0001 (Hi Word), Driver FFFB (Lo Word)

  Call_ATrap(0xA04E);     // Add Drive 1..."AddDrive" Trap (bill comment ---)

  // Add Drive 2
  put_word(SonyVars + kDrive2 + kTrack, 0x0000);     // Drive 2 at Track 0
  put_byte(SonyVars + kDrive2 + kWriteProt, 0x00);   // Drive 2 Writeable
  put_byte(SonyVars + kDrive2 + kDiskInPlace, 0x00); // Drive 2 No Disk
  put_byte(SonyVars + kDrive2 + kInstalled, 0x01);   // Drive 2 Installed
  put_byte(SonyVars + kDrive2 + kSides, 0xFF);       // Drive 2 Double Sided
  put_word(SonyVars + kDrive2 + kQDriveNo, 0x0002);  // Drive 2 is Drive 2
  put_word(SonyVars + kDrive2 + kQRefNum, 0xFFFB);   // Drive 2 uses .Sony
  put_byte(SonyVars + kDrive2 + kTwoSideFmt, 0x00);  // Drive 2 Single Format (Initially)
  put_byte(SonyVars + kDrive2 + kNewIntf, 0xFF);     // Drive 2 uses new interface
  put_word(SonyVars + kDrive2 + kDriveErrs, 0x0000); // Drive 2 has no errors
  
  m68k_areg(regs, 0) = SonyVars + kDrive2 + kQLink;
  m68k_dreg(regs, 0) = 0x0002FFFB; // Drive 2 0002 (Hi Word), Driver FFFB (Lo Word)

  Call_ATrap(0xA04E);     // Add Drive 2..."AddDrive" Trap (bill comment ---)
  
  // Add Drive 2
  put_word(SonyVars + kDrive3 + kTrack, 0x0000);     // Drive 2 at Track 0
  put_byte(SonyVars + kDrive3 + kWriteProt, 0x00);   // Drive 2 Writeable
  put_byte(SonyVars + kDrive3 + kDiskInPlace, 0x00); // Drive 2 No Disk
  put_byte(SonyVars + kDrive3 + kInstalled, 0x01);   // Drive 2 Installed
  put_byte(SonyVars + kDrive3 + kSides, 0xFF);       // Drive 2 Double Sided
  put_word(SonyVars + kDrive3 + kQDriveNo, 0x0003);  // Drive 2 is Drive 2
  put_word(SonyVars + kDrive3 + kQRefNum, 0xFFFB);   // Drive 2 uses .Sony
  put_byte(SonyVars + kDrive3 + kTwoSideFmt, 0x00);  // Drive 2 Single Format (Initially)
  put_byte(SonyVars + kDrive3 + kNewIntf, 0xFF);     // Drive 2 uses new interface
  put_word(SonyVars + kDrive3 + kDriveErrs, 0x0000); // Drive 2 has no errors
  
  m68k_areg(regs, 0) = SonyVars + kDrive3 + kQLink;
  m68k_dreg(regs, 0) = 0x0003FFFB; // Drive 2 0002 (Hi Word), Driver FFFB (Lo Word)

  Call_ATrap(0xA04E);     // Add Drive 2..."AddDrive" Trap (bill comment ---)

  m68k_areg(regs, 0) = tA0; // Restore Parameter Block and Device Control Entry
  m68k_dreg(regs, 0) = tD0;
  
  Sony_Insert(1, thePrefs.DrivePath[0]);
  Sony_Insert(2, thePrefs.DrivePath[1]);
  Sony_Insert(3, thePrefs.DrivePath[2]);
  
  Sony_Return(0x0000,0,1); // Immediate Return, No Error
}

// Sony_Prime
// Handles I/O to disks
void Sony_Prime (void)
{
  CPTR SonyVars = get_long(0x0134);
  ParamBlk = m68k_areg(regs, 0);
  DeviceCtl = m68k_areg(regs, 1);
  
  // See if the Drive is Installed
  if (Sony_Drive_Installed(get_word(ParamBlk + kioVRefNum)) == 0)
  {
    Sony_Return(0xFFC8,0,0); // No Such Drive (-56)
    return;
  }
  
  // See if a disk is currently inserted
  if (Sony_Disk_Inserted(get_word(ParamBlk + kioVRefNum)) == 0)
  {
    Sony_Return(0xFFBF,0,0); // Say it's offline (-65)
    return;
  }
  
  if (get_word(ParamBlk + kioVRefNum) == 0x0001)
    put_byte(SonyVars + kDrive1 + kDiskInPlace, 0x02); // Clamp Drive 1
  
  if (get_word(ParamBlk + kioVRefNum) == 0x0002)
    put_byte(SonyVars + kDrive2 + kDiskInPlace, 0x02); // Clamp Drive 2

  if (get_word(ParamBlk + kioVRefNum) == 0x0003)
    put_byte(SonyVars + kDrive3 + kDiskInPlace, 0x02); // Clamp Drive 2
  
  if ((get_word(ParamBlk + kioTrap) & 0xF0FF) == 0xA003) // PBWrite Trap
  {
    #ifdef _Sony_Writable
    Sony_Access(get_word(ParamBlk + kioVRefNum)); // Prepare to write a disk
    Sony_Return(0x0000,0,0);
    #else
    // Don't enable writes
    Sony_Return(0xFFD4,0,0); // Write Protected (-44)
    #endif
  }
  
  if ((get_word(ParamBlk + kioTrap) & 0xF0FF) == 0xA002) // PBRead Trap
  {
    Sony_Access(get_word(ParamBlk + kioVRefNum)); // Prepare to read a disk
    Sony_Return(0x0000,0,0);
  }
}
   
// Sony_Control
// Implements control csCodes for the Sony driver
void Sony_Control (void)
{
  UWORD csCode = 0;
  
  ParamBlk = m68k_areg(regs, 0);
  DeviceCtl = m68k_areg(regs, 1);
  
  if (Sony_Drive_Installed(get_word(ParamBlk + kioVRefNum)) == 0)
  {
    Sony_Return(0xFFC8,0,0); // No Such Drive (-56)
    return;
  }
  
  csCode = get_word(ParamBlk + kcsCode);
  
  switch (csCode)
  {
    case kKillIO : // Immediate Return of -1 (-17 in PowerMacs?)
    {
      Sony_Return(0xFFFF,0,1);
      break;
    }
    case kVerifyDisk :
    {
      if (Sony_Disk_Inserted(get_word(ParamBlk + kioVRefNum)) == 0)
      {
        Sony_Return(0xFFBF,0,0); // Say it's offline (-65)
        return;
      }
      Sony_Return(0x0000,0,0); // No Error (0)
      break;
    }
    case kEjectDisk :
    {
      Sony_Eject(get_word(ParamBlk + kioVRefNum));
      Sony_Return(0x0000,0,0); // No error (0)
      break;
    }
    case kFormatDisk :
    {
      if (Sony_Disk_Inserted(get_word(ParamBlk + kioVRefNum)) == 0)
      {
        Sony_Return(0xFFBF,0,0); // Say it's offline (-65)
        return;
      }
      Sony_Return(0xFFD4,0,0); // Write Protected (-44)
      return;
      break;
    }
    case kReturnDriveInfo :
    {
      if (get_word(ParamBlk + kioVRefNum) == 0x0001)
      {
        put_long(ParamBlk + kcsParam, 0x00000003); // 800K, Primary Internal Drive
        Sony_Return(0x0000,0,0); // No error (0)
        return;
      }
      if (get_word(ParamBlk + kioVRefNum) == 0x0002)
      {
        put_long(ParamBlk + kcsParam, 0x00000093); // 800K, Secondary External Drive
        Sony_Return(0x0000,0,0); // No error (0)
        return;
      }
      if (get_word(ParamBlk + kioVRefNum) == 0x0003)
      {
        put_long(ParamBlk + kcsParam, 0x00000093); // 800K, Secondary External Drive
        Sony_Return(0x0000,0,0); // No error (0)
        return;
      }
      break;
    }
    default :
    {
      Sony_Return(0xFFEF,0,0); // Driver can't respond to this Control call (-17)
      break;
    }
  }
}

// Sony_Status
// Handles the DriveStatus call
void Sony_Status (void)
{
  CPTR SonyVars = get_long(0x0134);
  CPTR Src = 0; CPTR Dst = 0;
  int Counter = 0;
  ParamBlk = m68k_areg(regs, 0);
  DeviceCtl = m68k_areg(regs, 1);

  if (get_word(ParamBlk + kcsCode) == kDriveStatus) // Drive Status Call
  {
    if (Sony_Drive_Installed(get_word(ParamBlk + kioVRefNum)) == 0)
    {
      Sony_Return(0xFFC8,0,0); // No Such Drive (-56)
      return;
    }
    if (get_word(ParamBlk + kioVRefNum) == 0x0001) // Drive 1 Status
    {
      Src = SonyVars + kDrive1;
      Dst = ParamBlk + kcsParam; 
      for (Counter = 1; Counter <= 11; Counter++)
      {
        put_word(Dst,get_word(Src));
        Src += 2; Dst += 2;
      }
      Sony_Return(0x0000,0,0);
      return;
    }
    if (get_word(ParamBlk + kioVRefNum) == 0x0002) // Drive 2 Status
    {
      Src = SonyVars + kDrive2;
      Dst = ParamBlk + kcsParam; 
      for (Counter = 1; Counter <= 11; Counter++)
      {
        put_word(Dst,get_word(Src));
        Src += 2; Dst += 2;
      }
      Sony_Return(0x0000,0,0);
      return;
    }
    if (get_word(ParamBlk + kioVRefNum) == 0x0003) // Drive 3 Status
    {
      Src = SonyVars + kDrive3;
      Dst = ParamBlk + kcsParam; 
      for (Counter = 1; Counter <= 11; Counter++)
      {
        put_word(Dst,get_word(Src));
        Src += 2; Dst += 2;
      }
      Sony_Return(0x0000,0,0);
      return;
    }
  }
  else
    Sony_Return(0xFFEE,0,0); // Can't Respond to Status Call (-18)
}

// Sony_Close
// Returns immediate error
void Sony_Close(void)
{ 
  Sony_Return(0xFFE8,0,1); // Can't Close Driver (-24)
}
