 /* 
  * UAE - The Un*x Amiga Emulator
  * 
  * memory management
  * 
  * Copyright 1995 Bernd Schmidt
  */

/* Enabling this adds one additional native memory reference per 68k memory
 * access, but saves one shift (on the x86). Enabling this is probably
 * better for the cache. My favourite benchmark (PP2) doesn't show a
 * difference, so I leave this enabled. */

#if 1 || defined SAVE_MEMORY
#define SAVE_MEMORY_BANKS
#endif

typedef ULONG (*mem_get_func)(CPTR) REGPARAM;
typedef void (*mem_put_func)(CPTR, ULONG) REGPARAM;
typedef UBYTE *(*xlate_func)(CPTR) REGPARAM;
typedef int (*check_func)(CPTR, ULONG) REGPARAM;

extern char *address_space, *good_address_map;
extern UBYTE *RAMmemory;

#undef DIRECT_MEMFUNCS_SUCCESSFUL
#include "../CPU/machdep/memory.h"

typedef struct {
    /* These ones should be self-explanatory... */
    mem_get_func lget, wget, bget;
    mem_put_func lput, wput, bput;
    /* Use xlateaddr to translate an Amiga address to a UBYTE * that can
     * be used to address memory without calling the wget/wput functions. 
     * This doesn't work for all memory banks, so this function may call
     * abort(). */
    xlate_func xlateaddr;
    /* To prevent calls to abort(), use check before calling xlateaddr.
     * It checks not only that the memory bank can do xlateaddr, but also
     * that the pointer points to an area of at least the specified size. 
     * This is used for example to translate bitplane pointers in custom.c */
    check_func check;
} addrbank;

extern addrbank RAMmem_bank;
extern addrbank ROMmem_bank;
extern ULONG RAMmem_size;

/* Default memory access functions */

extern int default_check(CPTR addr, ULONG size) REGPARAM;
extern UBYTE *default_xlate(CPTR addr) REGPARAM;

#define bankindex(addr) (((CPTR)(addr)) >> 16)

#ifdef SAVE_MEMORY_BANKS
extern addrbank **mem_banks;
#define get_mem_bank(addr) (*mem_banks[bankindex(addr)])
#define put_mem_bank(addr, b) (mem_banks[bankindex(addr)] = (b))
#else
extern addrbank mem_banks[65536];
#define get_mem_bank(addr) (mem_banks[bankindex(addr)])
#define put_mem_bank(addr, b) (mem_banks[bankindex(addr)] = *(b))
#endif

#if defined(NO_EXCEPTION_3) || CPU_LEVEL > 1
#define check_addr(a) 1
#else
#define check_addr(a) (((a) & 1) == 0)
#endif

extern int buserr;
extern UWORD last_op_for_exception_3;
extern CPTR last_addr_for_exception_3;
extern CPTR last_fault_for_exception_3;
extern CPTR m68k_currentpc(void);

extern void memory_init(void);    
extern void map_banks(addrbank *bank, int first, int count);

#ifndef NO_INLINE_MEMORY_ACCESS

#define longget(addr) (call_mem_get_func(get_mem_bank(addr).lget, addr))
#define wordget(addr) (call_mem_get_func(get_mem_bank(addr).wget, addr))
#define byteget(addr) (call_mem_get_func(get_mem_bank(addr).bget, addr))
#define longput(addr,l) (call_mem_put_func(get_mem_bank(addr).lput, addr, l))
#define wordput(addr,w) (call_mem_put_func(get_mem_bank(addr).wput, addr, w))
#define byteput(addr,b) (call_mem_put_func(get_mem_bank(addr).bput, addr, b))

#else

extern ULONG alongget(CPTR addr);
extern ULONG awordget(CPTR addr);
extern ULONG longget(CPTR addr);
extern ULONG wordget(CPTR addr);
extern ULONG byteget(CPTR addr);
extern void longput(CPTR addr, ULONG l);
extern void wordput(CPTR addr, ULONG w);
extern void byteput(CPTR addr, ULONG b);

#endif

#ifndef MD_HAVE_MEM_1_FUNCS

#define longget_1 longget
#define wordget_1 wordget
#define byteget_1 byteget
#define longput_1 longput
#define wordput_1 wordput
#define byteput_1 byteput

#endif

static __inline__ ULONG get_long(CPTR addr) 
{
    if (check_addr(addr))
	return longget_1(addr);
    buserr = 1;
    last_fault_for_exception_3 = addr;
    last_addr_for_exception_3 = m68k_currentpc();
    return 0;
}
static __inline__ ULONG get_word(CPTR addr) 
{
    if (check_addr(addr))
	return wordget_1(addr);
    buserr = 1;
    last_fault_for_exception_3 = addr;
    last_addr_for_exception_3 = m68k_currentpc();
    return 0;
}
static __inline__ ULONG get_byte(CPTR addr) 
{
    return byteget_1(addr); 
}
static __inline__ void put_long(CPTR addr, ULONG l) 
{
    if (!check_addr(addr)) {
	buserr = 1;
	last_fault_for_exception_3 = addr;
        last_addr_for_exception_3 = m68k_currentpc();
	return;
    }
    longput_1(addr, l);
}
static __inline__ void put_word(CPTR addr, ULONG w)
{
    if (!check_addr(addr)) {
	buserr = 1;
	last_fault_for_exception_3 = addr;
	last_addr_for_exception_3 = m68k_currentpc();
	return;
    }
    wordput_1(addr, w);
}
static __inline__ void put_byte(CPTR addr, ULONG b) 
{
    byteput_1(addr, b);
}

static __inline__ UBYTE *get_real_address(CPTR addr)
{
    return get_mem_bank(addr).xlateaddr(addr);
}

static __inline__ int valid_address(CPTR addr, ULONG size)
{
    return get_mem_bank(addr).check(addr, size);
}

extern void set_uae_vmac_address_space(void);
extern void memory_management_overlay(int);
extern void memory_init_uae_069(void);
extern void ROMmem_write_enable_set(unsigned int);

//IRIX CC needs a little new-line at the end of include files.

