 /* 
  * UAE - The Un*x Amiga Emulator
  * 
  * MC68000 emulation
  *
  * Copyright 1995 Bernd Schmidt
  */

extern int areg_byteinc[];
extern int imm8_table[];

extern int movem_index1[256];
extern int movem_index2[256];
extern int movem_next[256];

extern int fpp_movem_index1[256];
extern int fpp_movem_index2[256];
extern int fpp_movem_next[256];

extern int broken_in;

#define REGPARAM

typedef void cpuop_func(ULONG) REGPARAM;

struct cputbl {
    cpuop_func *handler;
    int specific;
    UWORD opcode;
};

extern void op_illg(ULONG) REGPARAM;

typedef char flagtype;

/* Arrrghh.. */
#if !defined NO_PREFETCH_BUFFER && !defined USE_POINTER
#define USE_POINTER
#endif
#if defined USE_COMPILER && !defined USE_POINTER
#define USE_POINTER
#endif

extern struct regstruct 
{
    ULONG regs[16];
    CPTR  usp,isp,msp;
    UWORD sr;
    flagtype t1;
    flagtype t0;
    flagtype s;
    flagtype m;
    flagtype x;
    flagtype stopped;
    int intmask;
    ULONG pc;
#ifdef USE_POINTER
    UBYTE *pc_p;
    UBYTE *pc_oldp;
#endif
    
    ULONG vbr,sfc,dfc;

    double fp[8];
    ULONG fpcr,fpsr,fpiar;

    ULONG spcflags;
    ULONG kick_mask;
#if !defined NO_PREFETCH_BUFFER
    /* Fellow sources say this is 4 longwords. Trust Petter... */
    UBYTE prefetch[16];
    UBYTE *prefetch_pos;
#endif
} regs, lastint_regs;

#if !defined NO_PREFETCH_BUFFER
#endif

#define m68k_dreg(r,num) ((r).regs[(num)])
#define m68k_areg(r,num) ((r).regs[(num)+8])

#ifndef NO_PREFETCH_BUFFER
extern void fill_prefetch(void);
extern ULONG nextibyte(void);
extern ULONG nextiword(void);
extern ULONG nextilong(void);

#else

#define fill_prefetch() do { (void)0; } while (0)


// bill comment ---
// Damn Damn Damn Damn !!!!
//
// "regs.pc_p" is defined as a UBYTE, which
// is 8 bits wide.
//
// In the old UAE 0.60 emulator, it was defined
// as UWORD, which is a 16 wide entity.
// Since this version of nextilong() is from the old emulator,
// the assignments in this function were only
// copying 8 bits of data instead of 16...
//
// I must make a cast to a 16 entity, so that upon
// dereferencing, it will copy a 16 bits instead of 8. 
//
// All auto increment operations must also be changed
// to explicit specified human defined constants, because
// the compiler will vary the autoincrement byte size count
// depending on size of the base type. This is what
// led to the incorrect functioning of nextilong(),
// nextibyte(), nextiword()...
//

#ifdef USE_POINTER

static __inline__ ULONG nextibyte(void)
{
//    ULONG r = do_get_mem_byte(regs.pc_p+1);
    ULONG r;

//debugout(true, "nextibyte = %02x\n", (UBYTE) regs.pc_p);
    r = do_get_mem_byte(regs.pc_p+1);
    regs.pc_p += 2;
    return r;
}

static __inline__ ULONG nextiword(void)
{
//    ULONG r = do_get_mem_word((UWORD *)regs.pc_p);
    ULONG r;

//debugout(true, "nextiword = %04hx\n", (UWORD) regs.pc_p);
    r = do_get_mem_word((UWORD *)regs.pc_p);
    regs.pc_p += 2;
    return r;
}

static __inline__ ULONG nextilong(void)
{
//    ULONG r = do_get_mem_long((ULONG *)regs.pc_p);
    ULONG r;
//debugout(true, "nextilong = %08lx\n", (ULONG) regs.pc_p);
    r = do_get_mem_long((ULONG *)regs.pc_p);
    regs.pc_p += 4;
    return r;
}

#else

static __inline__ ULONG nextibyte(void)
{
    ULONG r = get_byte(regs.pc+1);
    regs.pc += 2;
    return r;
}

static __inline__ ULONG nextiword(void)
{
    ULONG r = get_word(regs.pc);
    regs.pc += 2;
    return r;
}

static __inline__ ULONG nextilong(void)
{
    ULONG r = get_long(regs.pc);
    regs.pc += 4;
    return r;
}

#endif
#endif

#ifdef USE_POINTER

#if !defined(USE_COMPILER)
static __inline__ void m68k_setpc(CPTR newpc)
{
    regs.pc_p = regs.pc_oldp = (UBYTE *) get_real_address(newpc);
    regs.pc = newpc;
#if USER_PROGRAMS_BEHAVE > 0
    if ((newpc & 0xF80000) != regs.kick_mask)
	regs.spcflags |= SPCFLAG_MODE_CHANGE;
    regs.kick_mask = newpc & 0xF80000;
#endif
    fill_prefetch();
}
#else
extern void m68k_setpc(CPTR newpc);
#endif

static __inline__ CPTR m68k_getpc(void)
{
    return regs.pc + ((char *)regs.pc_p - (char *)regs.pc_oldp);
}

#else

static __inline__ void m68k_setpc(CPTR newpc)
{
//    regs.pc = newpc;
// bill mod
    regs.pc = newpc & 0x00ffFFFF;
}

static __inline__ CPTR m68k_getpc(void)
{
    return regs.pc;
}
#endif

#ifdef USE_COMPILER
extern void m68k_setpc_fast(CPTR newpc);
extern void m68k_setpc_bcc(CPTR newpc);
extern void m68k_setpc_rte(CPTR newpc);
#else
#define m68k_setpc_fast m68k_setpc
#define m68k_setpc_bcc  m68k_setpc
#define m68k_setpc_rte  m68k_setpc
#endif

static __inline__ void m68k_setstopped(int stop)
{
    regs.stopped = stop;
    if (stop)
	regs.spcflags |= SPCFLAG_STOP;
}

#if CPU_LEVEL > 1
static __inline__ ULONG get_disp_ea (ULONG base)
{
    UWORD dp = nextiword();
    int reg = (dp >> 12) & 15;
    LONG regd = regs.regs[reg];
    if ((dp & 0x800) == 0)
	regd = (LONG)(WORD)regd;
    regd <<= (dp >> 9) & 3;
    if (dp & 0x100) {
	LONG outer = 0;
	if (dp & 0x80) base = 0;
	if (dp & 0x40) regd = 0;

	if ((dp & 0x30) == 0x20) base += (LONG)(WORD)nextiword();
	if ((dp & 0x30) == 0x30) base += nextilong();
	
	if ((dp & 0x3) == 0x2) outer = (LONG)(WORD)nextiword();
	if ((dp & 0x3) == 0x3) outer = nextilong();
	
	if ((dp & 0x4) == 0) base += regd;
	if (dp & 0x3) base = get_long (base);
	if (dp & 0x4) base += regd;
	
	return base + outer;
    } else {
	return base + (LONG)((BYTE)dp) + regd;
    }
}
#else
static __inline__ ULONG get_disp_ea (ULONG base)
{
    UWORD dp = nextiword();
    int reg = (dp >> 12) & 15;
    LONG regd = regs.regs[reg];
    if ((dp & 0x800) == 0)
	regd = (LONG)(WORD)regd;
    return base + (BYTE)(dp) + regd;
}
#endif

extern LONG ShowEA(int reg, amodes mode, wordsizes size, char *buf);

extern void MakeSR(void);
extern void MakeFromSR(void);
extern void Exception(int, CPTR);
extern void dump_counts(void);
extern void m68k_move2c(int, ULONG *);
extern void m68k_movec2(int, ULONG *);
extern void m68k_divl (ULONG, ULONG, UWORD, CPTR);
extern void m68k_mull (ULONG, ULONG, UWORD);
extern void init_m68k (void);
extern void m68k_go(int);
extern void m68k_dumpstate(CPTR *);
extern void m68k_disasm(CPTR,CPTR *,int);
extern void m68k_reset(void);

extern void mmu_op (ULONG, UWORD);

extern void fpp_opp (ULONG, UWORD);
extern void fdbcc_opp (ULONG, UWORD);
extern void fscc_opp (ULONG, UWORD);
extern void ftrapcc_opp (ULONG,CPTR);
extern void fbcc_opp (ULONG, CPTR, ULONG);
extern void fsave_opp (ULONG);
extern void frestore_opp (ULONG);

#ifdef MEMFUNCS_DIRECT_REQUESTED
#define CPU_OP_NAME(a) op_direct ## a
#else
#define CPU_OP_NAME(a) op ## a
#endif

extern struct cputbl op_smalltbl[];
extern struct cputbl op_direct_smalltbl[];

extern cpuop_func *cpufunctbl[65536] ASM_SYM_FOR_FUNC ("cpufunctbl");
#if USER_PROGRAMS_BEHAVE > 0
extern cpuop_func *cpufunctbl_behaved[65536] ASM_SYM_FOR_FUNC ("cpufunctbl_behaved");
#endif

//IRIX CC needs a little new-line at the end of include files.

