/*
 * vMac - SVGA-Linux.c
 * Linux SVGA Interface for vMac
 * Copyright 1997 Matt Gischer (matt@mlmicro.com) and 
 * Phil Rutschman (machack@rsnsoft.com)
 */

/*
 * Modifications by Michael Hanni (1998) (mhanni@sprintmail.com)
 */
 

#include "sysconfig.h"
#include "sysdeps.h"
#include "Screen.h"
#include "Sony.h"
#include "RAM.h"
#include "VIA.h"
#include "prefs.h"
#include "Keyboard.h"
#include "Mouse.h"

#include <ncurses.h>
#include <vga.h>
#include <vgakeyboard.h>
#include <vgamouse.h>
#include <stdio.h>

UBYTE vPage2 = 1;
UBYTE Draw_Count = 0;		/* PC - Counter for Drawing Screen */

volatile int buttonstate;
volatile int mouse_x, mouse_y;
volatile int lastmx, lastmy;

static int macDisplayHeight = 342;
static int macDisplayWidth = 512;
int oldscan, oldstate;
int VIDMODE;
int KeyState[256];

int Shift, Command, Option, CapsLock;

void kb_handle(int ScanCode, int NewState);
void Quit_vMac(void);

#define SC_ESC 0xFFFE
#define SC_CTR 0xFFFD
#define SC_ALT 0xFFFC
#define SC_SFT 0xFFFB
#define SC_CLK 0xFFFA

int MacKeys[128] =
{
    0xFFFF, SC_ESC, 0x1231, 0x1332, 0x1433, 0x1534, 0x1735, 0x1636,
    0x1A37, 0x1C38, 0x1939, 0x1D30, 0x1B2D, 0x183D, 0x3308, 0x3009,
    0x0C71, 0x0D77, 0x0E65, 0x0F72, 0x1174, 0x1079, 0x2075, 0x2269,
    0x1F6F, 0x2370, 0xFFFF, 0xFFFF, 0x490D, SC_CTR, 0x0061, 0x0173,
    0x0264, 0x0366, 0x0567, 0x0468, 0x266A, 0x286B, 0x256C, 0x293B,
    0x2727, 0xB27E, SC_SFT, 0x2A5C, 0x067A, 0x0778, 0x0863, 0x0976,
    0x0B62, 0x2D6E, 0x2E6D, 0x2B2C, 0x2F2E, 0x2C2F, SC_SFT, 0x9C2A,
    SC_ALT, 0x3120, SC_CLK, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF,
    0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0x1A37,
    0x1C38, 0x1939, 0x1B2D, 0x1534, 0x1735, 0x1636, 0x982B, 0x1231,
    0x1332, 0x1433, 0x1D30, 0x2F2E, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF,
    0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF,
    0x490D, SC_CTR, 0x2C2F, 0xFFFF, SC_ALT, 0xFFFF, 0xFFFF, 0x4D1E,
    0xFFFF, 0x461C, 0x421D, 0xFFFF, 0x481F, 0xFFFF, 0xFFFF, 0xFFFF,
    0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF,
    0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF
};

unsigned char EscapeKeys[128] =
{
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 96, 97, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 98, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 99,
    100, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 101, 102,
    103, 104, 0, 105, 0, 106, 0, 107,
    108, 109, 110, 111, 0, 0, 0, 0,
    0, 0, 0, 125, 126, 127, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
};

int Screen_Init(int argc, char** argv)
{
    int y;
    VIDMODE = 10;

    if (vga_setmode(VIDMODE) < 0) {
	sleep(1);
	vga_setmode(TEXT);
	fprintf(stderr, "SVGAlib doesn't like my video mode. Giving up.\n");
	return 0;
    }
    vga_setmousesupport(1);
    if (keyboard_init() != 0) {
	vga_setmode(TEXT);
	fprintf(stderr, "Are you sure you have a keyboard??\n");
	return 0;
    }
    keyboard_seteventhandler(kb_handle);
    if ((mouse_init("/dev/mouse", vga_getmousetype(), 10)) != 0) {
	printf("Mouse could not be initialized\n");
    }
    mouse_setxrange(64, 576);
    mouse_setyrange(69, 411);
    mouse_setposition(74, 80);

    vga_setpalette(0, 255, 255, 255);
    vga_setpalette(1, 0, 0, 0);
    vga_setcolor(1);
    for (y = 0; y < 480; y++) {
	vga_drawline(0, y, 640, y);
    }
    thePrefs.FrameRate = 1; 
    printf("Drawing Every %i frames\n", thePrefs.FrameRate);

    return 0;
}

void Screen_Draw(void)
{
    int x, y, x2;
    unsigned char *buf;
    int scanline = 0;
    ULONG *baseaddr;
#define BUFLEN 512
    unsigned char cvtbuf[BUFLEN];

    if (vPage2 == 1) {
	buf = (unsigned char *) RAM_Xlate(kMain_Buffer);
    } else {
	buf = (unsigned char *) RAM_Xlate(kAlternate_Buffer);
    }
    (ULONG)baseaddr = RAM_GetB(kMain_Buffer);
//    if ((Draw_Count % thePrefs.FrameRate) == 0) {
	for (y = 0; y < 342; y++, buf += 64, scanline += 80) {
	    /* Bytesex and inversion */
	    for (x = x2 = 0; x < BUFLEN; x += 16) {
		cvtbuf[x + 0] = ((buf[x2] & 0x80)>> 7);
		cvtbuf[x + 1] = ((buf[x2] & 0x40) >> 6);
		cvtbuf[x + 2] = ((buf[x2] & 0x20) >> 5);
		cvtbuf[x + 3] = ((buf[x2] & 0x10) >> 4);
		cvtbuf[x + 4] = ((buf[x2] & 0x08) >> 3);
		cvtbuf[x + 5] = ((buf[x2] & 0x04) >> 2);
		cvtbuf[x + 6] = ((buf[x2] & 0x02) >> 1);
		cvtbuf[x + 7] = ((buf[x2++] & 0x01));
		cvtbuf[x + 8] = ((buf[x2] & 0x80) >> 7);
		cvtbuf[x + 9] = ((buf[x2] & 0x40) >> 6);
		cvtbuf[x + 10] = ((buf[x2] & 0x20) >> 5);
		cvtbuf[x + 11] = ((buf[x2] & 0x10) >> 4);
		cvtbuf[x + 12] = ((buf[x2] & 0x08) >> 3);
		cvtbuf[x + 13] = ((buf[x2] & 0x04) >> 2);
		cvtbuf[x + 14] = ((buf[x2] & 0x02) >> 1);
		cvtbuf[x + 15] = ((buf[x2++] & 0x01));
	    }
	    vga_drawscansegment((unsigned char *) cvtbuf, 64, 69 + y, BUFLEN);
	}
//    }
//    Draw_Count++;
//    Catch_SVGA_Events();
}


void Screen_Reset()
{
    printf("Screen Reset\n");
    return;			/* need to write refresh, but first need to get Draw to work */
}


void Process_Screen_Events()
{
    int x, y;
    if (mouse_update()) {
	lastmx = mouse_getx() - 64;
	lastmy = mouse_gety() - 69;
	buttonstate = mouse_getbutton();
    }
    keyboard_update();
}

void kb_handle(int ScanCode, int NewState)
{
    int MacKey, Mods = 0, Escape = 0;

    MacKey = MacKeys[ScanCode];
    if(EscapeKeys[ScanCode]!=0) Escape=1; else Escape = 0;

    if (Shift) {
	Mods = 0x0200;
	MacKey = MacKey - 32;	/* Lowercase -> Uppercase in ASCII */
	Shift = 0;
    } else if (Command) {
	Mods = 0x0100;
	Command = 0;
    } else if (CapsLock) {
	Mods = 0x0400;
	CapsLock = 0;
    } else if (Option) {
	Mods = 0x0800;
	Option = 0;
    }

    if ((oldscan == ScanCode) && (oldstate == NewState)) {
	PC_PostEvent(0x005, (ScanCode << 8) + MacKey, Mods);
    }
    oldscan = ScanCode;
    oldstate = NewState;

    switch (ScanCode) {
    case 42: /* Shift Keys */
    case 54:
	Shift = 1;
	break;
    case 56:  /* Command Keys */
    case 100: 
	Command = 1;
	break;
    case 58:  /* Caps Lock */
	Mods = 0x0400;
	CapsLock = 1;
	break;
    case 29: /* Option Keys */
    case 97:
	Option = 1;
	break;
    case 87:			/* F11 */
	Sony_Eject(2);
	Sony_Insert(2, "/dev/fd0");
	break;
    case 88:			/* F12 */
	Quit_vMac();
	break;
    }

/* Return now so we don't print out control keys */
    if (Shift || Command || Option || CapsLock || Escape) {
	oldstate = -1;		/* Make it so we can't repeat */
     return;
    }
    MacKey &= 0x7F;
    if (NewState == 0)
	PC_PostEvent(0x004, ((ScanCode << 8) + MacKey), Mods);
    if (NewState == 1)
	PC_PostEvent(0x003, ((ScanCode << 8) + MacKey), Mods);
}


// VIA Interface Functions

UBYTE
VIA_GORA6(void)			// Main/Alternate Screen Buffer
 {
#ifdef _VIA_Interface_Debug
    printf("VIA ORA6 attempts to be an input\n");
#endif
    return 0;
}

void VIA_PORA6(UBYTE Data)
{
    vPage2 = Data;
}

UBYTE
VIA_GORB6(void)			// Video Beam in Display
 {
    return 0;			// Assume it is

}

void VIA_PORB6(UBYTE Data)
{
#ifdef _VIA_Interface_Debug
    printf("VIA ORB6 attempts to be an output\n");
#endif
}
