#import "MacView.h"
#import "global.h"

extern void Keyboard_Down(ULONG Key);
extern void Keyboard_Up(ULONG Key);
extern void PC_PostEvent(UWORD What, ULONG Msg, UWORD Mods);

@implementation MacView

- initFrame:(const NXRect *)rect
{
    // Initialize the view
    self = [super initFrame:rect];

    // Create NXBitmapImageRep
    macImage1 = macImage2 = nil;

    theMacView = self;
    return self;
}

// free method to free all the images along with the view.

- free
{
    if( macImage1 != nil ) [macImage1 free];
    if( macImage2 != nil ) [macImage2 free];
    return [super free];
}

- generateMacImage:(unsigned char *)screen
{
    id image = [[NXBitmapImageRep alloc] initData:screen
                                         pixelsWide:macDisplayWidth
                                         pixelsHigh:macDisplayHeight
                                         bitsPerSample:1
                                         samplesPerPixel:1
                                         hasAlpha:NO
                                         isPlanar:YES
                                         colorSpace:NX_OneIsBlackColorSpace
                                         bytesPerRow:(macDisplayWidth + 7)/8
                                         bitsPerPixel:1];
    return image;
}

- generateMacImages
{
    macImage1 = [self generateMacImage:(unsigned char *)RAM_Xlate(kMain_Buffer)];
    macImage2 = [self generateMacImage:(unsigned char *)RAM_Xlate(kAlternate_Buffer)];
    return self;
}

// drawSelf:: simply redisplays the contents of the view. The source and
// destination rectangles are updated from the bitmaps while the result
// rectangle is created by compositing the two bitmaps.

- drawSelf:(NXRect *)r :(int) count
{
    [self lockFocus];
    if( vPage2 == 1 )
    {
        [macImage1 draw];
    }
    else
    {
        [macImage2 draw];
    }
    [self unlockFocus];
    return self;
}

- (BOOL)acceptsFirstResponder
{
    return YES;
}

- becomeKeyWindow
{
    [(MacTimer *)timer resume];
	return self ;
}

- resignKeyWindow
{
    [timer suspend];
	return self ;
}

- mouseUp:(NXEvent *)theEvent
{
    [self mouseMoved:theEvent];
    buttonstate = 0;
    return self;
}

- mouseDown:(NXEvent *)theEvent
{
    [self mouseMoved:theEvent];
    buttonstate = 1;
    return self;
}

- mouseMoved:(NXEvent *)theEvent
{
    NXPoint mouseLoc;

    mouseLoc.x = theEvent->location.x;
    mouseLoc.y = theEvent->location.y;
	[self convertPoint:&mouseLoc fromView:nil];

    lastmx = mouseLoc.x;
    lastmy = macDisplayHeight - mouseLoc.y;

    if( lastmx < 0 )                      lastmx = 0;
    else if( macDisplayWidth  <= lastmx ) lastmx = macDisplayWidth - 1;

    if( lastmy < 0 )                      lastmy = 0;
    else if( macDisplayHeight <= lastmy ) lastmy = macDisplayHeight - 1;
    return self;
}

- mouseEntered:(NXEvent *)theEvent
{
    return [self mouseMoved:theEvent];
}

- mouseDragged:(NXEvent *)theEvent
{
    return [self mouseMoved:theEvent];
}

#define LEFTARROW	172
#define RIGHTARROW	174
#define UPARROW		173
#define DOWNARROW	175

- keyUp:(NXEvent *)theEvent
{
    switch(theEvent->data.key.charCode)
    {
    case LEFTARROW:
        Keyboard_Up(0x01c);
        break;
    case RIGHTARROW:
        Keyboard_Up(0x01d);
        break;
    case UPARROW:
        Keyboard_Up(0x01e);
        break;
    case DOWNARROW:
        Keyboard_Up(0x01f);
        break;
    default:
        {
            extern UBYTE KeyCode_Table [128];

            unsigned short modifier = 0;
            unsigned int charCode = theEvent->data.key.charCode;
            unsigned int keyCode  = KeyCode_Table[charCode & 0xFF];

            if((keyCode & 0x80) == 0x80)
            {
                modifier = 0x0200; // Set Modifiers up (add any if required!)
                keyCode &= 0x7F;   // Strip Shift Indicator  
            }

			if(theEvent->flags & NX_COMMANDMASK)
				modifier |= 0x0100;
			if(theEvent->flags & NX_SHIFTMASK)
				modifier |= 0x0200;
			if(theEvent->flags & NX_CONTROLMASK)
				modifier |= 0x0400;
			if(theEvent->flags & NX_ALTERNATEMASK)
				modifier |= 0x0800;

            PC_PostEvent(0x0004, ((keyCode << 8) + (charCode & 0x7F)), modifier);
        }
    }
    return self;
}

- keyDown:(NXEvent *)theEvent
{
    switch(theEvent->data.key.charCode)
    {
    case LEFTARROW:
        Keyboard_Down(0x01c);
        break;
    case RIGHTARROW:
        Keyboard_Down(0x01d);
        break;
    case UPARROW:
        Keyboard_Down(0x01e);
        break;
    case DOWNARROW:
        Keyboard_Down(0x01f);
        break;
    default:
        {
            extern UBYTE KeyCode_Table [128];

            unsigned short modifier = 0;
            unsigned int charCode = theEvent->data.key.charCode;
            unsigned int keyCode  = KeyCode_Table[charCode & 0xFF];

            if((keyCode & 0x80) == 0x80)
            {
                modifier = 0x0200; // Set Modifiers up (add any if required!)
                keyCode &= 0x7F;   // Strip Shift Indicator  
            }

			if(theEvent->flags & NX_COMMANDMASK)
				modifier |= 0x0100;
			if(theEvent->flags & NX_SHIFTMASK)
				modifier |= 0x0200;
			if(theEvent->flags & NX_CONTROLMASK)
				modifier |= 0x0400;
			if(theEvent->flags & NX_ALTERNATEMASK)
				modifier |= 0x0800;

            PC_PostEvent(0x0003, ((keyCode << 8) + (charCode & 0x7F)), modifier);
        }
    }
    return self;
}

@end
