#import "vMacController.h"
#import "PrefController.h"
#import "global.h"

@implementation vMacController

- reset:sender
{
    mutex_lock(conditionLock);
    thread_condition = CTHREAD_RESET;
    mutex_unlock(conditionLock);
    return self;
}

- suspendResume:sender
{
    if( !suspend ) // suspend
    {
        [self appSuspend];
        [timer stopTimer];
        suspend = YES;
        [suspendMenuCell setTitle:"Resume"];
    }
    else // resume
    {
        [self appResume];
        [timer startTimer];
        suspend = NO;
        [suspendMenuCell setTitle:"Suspend"];
    }
    return self;
}

- appDidInit:sender
{
    // not so good.
    NXBundle *bundle = [NXBundle bundleForClass:[self class]];
    chdir([bundle directory]); // change current directory for application.

    // bring to front main window
    [[macView window] makeFirstResponder:macView];
    [[macView window] makeKeyAndOrderFront:self];
    [[macView window] addToEventMask:NX_MOUSEMOVEDMASK|NX_LMOUSEDRAGGEDMASK];

    // emulation thread does not run yet
    ct = NULL;

    [self changePreferences];
	vMac_Install_SignalHandler();
	init_m68k();

	if( custominit() == 0 )
	{
        extern void m68k_run_1(void);

        thread_condition = CTHREAD_NORMAL;
        conditionLock = mutex_alloc();
        done = condition_alloc();
        resume = condition_alloc();
        suspended = condition_alloc();
        diskInsert(0,NULL);

        [macView generateMacImages];
		m68k_reset();
        cthread_detach(ct = cthread_fork((cthread_fn_t)m68k_run_1,self));
        [(MacTimer *)[timer init] startTimer];
	}
	else
	{
		NXRunAlertPanel( "vMac Error",
                         "Sorry, vMac encountered errors and cannot continue. "
                         "Please check console messages.\n\n"
                         "If you find a message missing the vMac.ROM, "
                         "please place your Macintosh Plus rom image in the application "
                         "directory and name it \"vMac.ROM\".",
                         NULL, NULL, NULL );
        [NXApp terminate:nil];
	}
    return self;
}

- appWillTerminate:sender
{
    condition_signal(resume);
    mutex_lock(conditionLock);
    thread_condition = CTHREAD_QUIT;
    condition_wait(done, conditionLock);
    mutex_unlock(conditionLock);
    [timer free];
    vMac_Uninstall_SignalHandler();
    return self;
}

- openPreference:sender
{
    return self;
}

- changePreferences;
{
    // future work. currently ignore following parameters.
	thePrefs.NoXmouse=0;
	thePrefs.MacSound=0;

    // NEXTSTEP needs vMac.ROM in the application directory.
	thePrefs.CurrentROMPath=0;

	strcpy(thePrefs.DrivePath[0], [prefController systemDisk]);
	strcpy(thePrefs.DrivePath[1], "empty2.dsk");
	strcpy(thePrefs.DrivePath[2], "empty3.dsk");
	strcpy(thePrefs.ROMPath[0],"vMac.ROM");
	thePrefs.RAMSize = [prefController ramSize];

    [self appSuspend];
	thePrefs.NumInstructions = (int)(DEFAULT_NUM_INSTRUCTIONS/[prefController clockRate]);
    [self appResume];

    [timer setActiveInterval:[prefController refreshRate]];
    return self;
}

#define LOAD_FILE_1 (0)
#define MOUNT_FDD_1 (1)
#define LOAD_FILE_2 (2)
#define MOUNT_FDD_2 (3)

- loadfile:sender
{
    char *path;
    const char *const types[] = {"dsk","hfv", NULL};
    id openPanel = [OpenPanel new];
    unsigned short drive = 0;

    [openPanel allowMultipleFiles:NO];
    path = alloca(1 + MAXPATHLEN);

    switch([[sender selectedCell] tag])
    {
    case LOAD_FILE_1:
        [openPanel setTitle:"Open file for drive1"];
        drive = 0x0002;
        break;

    case LOAD_FILE_2:
        [openPanel setTitle:"Open file for drive2"];
        drive = 0x0003;
        break;
    }

    if( drive != 0x0000 && [openPanel runModalForDirectory:path file:NULL types:types])
    {
        const char *file = [openPanel filename];
        if( file != NULL)
        {
            diskInsert(drive, file);
        }
    }
    return self;
}

- mountfdd:sender
{
    unsigned short drive = ([[sender selectedCell] tag] == MOUNT_FDD_1) ? 0x0002 : 0x0003;
    diskInsert(drive, "/dev/rfd0b");
    return self;
}

- appSuspend
{
    if( ct != NULL )
    {
        mutex_lock(conditionLock);
        thread_condition = CTHREAD_STOP;
        condition_wait(suspended, conditionLock);
        mutex_unlock(conditionLock);
    }
    return self;
}

- appResume
{
    if( ct != NULL )
    {
        mutex_lock(conditionLock);
        thread_condition = CTHREAD_NORMAL;
        mutex_unlock(conditionLock);
        condition_signal(resume);
    }
    return self;
}

@end
