/*
 * vMac - Screen-Motif.c
 * Xt/Lesstif(Motif) X Window Interface.
 * Copyright 1997 Michael Hanni (mhanni@sprintmail.com)
 */

#include "sysconfig.h"
#include "sysdeps.h"
#include "config.h"
#include "options.h"
#include "custom.h"

#include "Screen.h"
#include "RAM.h"  
#include "VIA.h"
#include "Keyboard.h"
#include "Sony.h"
#include "Sound.h"
#include "Mouse.h"
#include "prefs.h"
#include "debug.h"
#include "universal.h"

#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>

#include "Screen-GTK+-Menus.h"

int 		lastmx, lastmy;
int 		buttonstate;
int          Key, Key2 = 0x01C, Key3 = 0x01D, Key4 = 0x01E, Key5 = 0x01F;
char         ourKey;  
UWORD        Modifier;

UBYTE vPage2 = 1;

extern void Quit_vMac(void);
extern void m68k_reset(void);

gint          	macDisplayHeight = 342;
gint          	macDisplayWidth = 512;
GtkWidget 	*window, *drawing, *widget, *file_window;
GdkImage 	*vmacimg;
GdkVisual	*gVisual;
GdkGC		*vmacgc;
gpointer 	*image_Mem1;
GdkColor 	fg, bg;

static void popup_unmap(GtkWidget *widget, gpointer data)
{
	gtk_widget_destroy(file_window);
	file_window = 0;
}

static void gdk_m68k_reset(GtkWidget *widget, gpointer data)
{
	m68k_reset();
	gtk_widget_destroy(file_window);
	file_window = 0;
}

static gint delete_event(GtkWidget *widget, GdkEvent *event, gpointer data)
{
         Quit_vMac ();

         return (FALSE); 
}

static void destroy (GtkWidget *widget, gpointer data)
{
         Quit_vMac ();
}

static gint
motion_notify_event (GtkWidget *widget, GdkEventMotion *event) {

       	int x, y;
       	GdkModifierType state;

       	if (event->is_hint) {
	   	gdk_window_get_pointer (event->window, &lastmx, &lastmy, &state);
       	}
       	else
       	{
           	lastmx = event->x;
           	lastmy = event->y;
	   	printf("x: %d  y: %d\n", lastmx, lastmy);
       	}

       	return TRUE;
}

static gint
key_press_event (GtkWidget *widget, GdkEventKey *event)
{
	int keyval;

        keyval = event->keyval;

        switch (keyval) {
                case GDK_Left:
                        Keyboard_Down(Key2);
                        break;
                case GDK_Right:
                        Keyboard_Down(Key3);
                        break;
                case GDK_Up:
                        Keyboard_Down(Key4);
                        break;
                case GDK_Down:
                        Keyboard_Down(Key5);
                        break;
                case GDK_Meta_L:
                case GDK_Meta_R:
                case GDK_Alt_L:
                case GDK_Alt_R:
                case GDK_Control_L:
                case GDK_Control_R:   
                case GDK_Shift_L:
                case GDK_Shift_R:
                case GDK_Caps_Lock:
                        break;
                default:
                        if (event->state & GDK_SHIFT_MASK)
                                Modifier = 0x0200;
                        else if (event->state & GDK_CONTROL_MASK)
                                Modifier = 0x0800;
                        else if (event->state & GDK_MOD1_MASK)
                                Modifier = 0x0100;
                        else if (event->state & GDK_LOCK_MASK)
                                Modifier = 0x0400;
                        else
                                Modifier = 0x0000;
                        PC_PostEvent (3, keyval, Modifier);
                        break;
        }

       	return TRUE;
}

static gint
key_release_event (GtkWidget *widget, GdkEventKey *event)
{
	int keyval;

        keyval = event->keyval;

        switch (keyval) {
                case GDK_Left:
                        Keyboard_Up(Key2);
                        break;
                case GDK_Right:
                        Keyboard_Up(Key3);
                        break;
                case GDK_Up:
                        Keyboard_Up(Key4);
                        break;
                case GDK_Down:
                        Keyboard_Up(Key5);
                        break;
                case GDK_Meta_L:
                case GDK_Meta_R:
                case GDK_Alt_L:
                case GDK_Alt_R:
                case GDK_Control_L:
                case GDK_Control_R:   
                case GDK_Shift_L:
                case GDK_Shift_R:
                case GDK_Caps_Lock:
                        break;
                default:
                        if (event->state & GDK_SHIFT_MASK)
                                Modifier = 0x0200;
                        else if (event->state & GDK_CONTROL_MASK)
                                Modifier = 0x0800;
                        else if (event->state & GDK_MOD1_MASK)
                                Modifier = 0x0100;
                        else if (event->state & GDK_LOCK_MASK)
                                Modifier = 0x0400;
                        else
                                Modifier = 0x0000;
                        PC_PostEvent (4, keyval, Modifier);
                        break;
        }

       	return TRUE;
}

static gint
button_press_event (GtkWidget *widget, GdkEventButton *event)
{
	buttonstate = 1;
       	return TRUE;
}

static gint
button_release_event (GtkWidget *widget, GdkEventButton *event)
{
	buttonstate = 0;
       	return TRUE;
}

int Screen_Init(int argc, char **argv) {

	GtkWidget *menubar, *aBox;
	GtkAcceleratorTable *accel;

	gtk_init (&argc, &argv);

	window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW (window), "vMac Unix/X GTK+ Alpha-1");

    	gtk_signal_connect (GTK_OBJECT (window), "destroy",
                             GTK_SIGNAL_FUNC (destroy), NULL);
    	gtk_signal_connect (GTK_OBJECT (window), "delete_event",
                             GTK_SIGNAL_FUNC (delete_event), NULL);

	aBox = gtk_vbox_new (FALSE, FALSE);
        gtk_widget_show (aBox);

	gtk_container_add (GTK_CONTAINER (window), aBox);

	get_main_menu(&menubar, &accel);
	gtk_window_add_accelerator_table(GTK_WINDOW(window), accel);
	gtk_box_pack_start(GTK_BOX(aBox), menubar, FALSE, TRUE, 0);
        gtk_widget_show(menubar);

	drawing = gtk_drawing_area_new ();
        gtk_drawing_area_size (GTK_DRAWING_AREA (drawing),
                               macDisplayWidth,
                               macDisplayHeight);

        gtk_box_pack_start_defaults (GTK_BOX(aBox), drawing);

/* pack_end */

        gtk_widget_show (drawing);

  	gtk_signal_connect (GTK_OBJECT (drawing), "motion_notify_event",
                           (GtkSignalFunc) motion_notify_event, NULL);
       	gtk_signal_connect (GTK_OBJECT (drawing), "button_press_event",
                           (GtkSignalFunc) button_press_event, NULL);
       	gtk_signal_connect (GTK_OBJECT (drawing), "button_release_event",
                           (GtkSignalFunc) button_release_event, NULL);
       	gtk_signal_connect (GTK_OBJECT (drawing), "key_press_event",
                           (GtkSignalFunc) key_press_event, NULL);
       	gtk_signal_connect (GTK_OBJECT (drawing), "key_release_event",
                           (GtkSignalFunc) key_release_event, NULL);

	gtk_widget_set_events (drawing, GDK_BUTTON_PRESS_MASK
			      | GDK_KEY_PRESS_MASK
			      | GDK_KEY_RELEASE_MASK
			      | GDK_BUTTON_RELEASE_MASK
                              | GDK_POINTER_MOTION_MASK
                              | GDK_POINTER_MOTION_HINT_MASK);

	image_Mem1 = calloc(1, 64 * macDisplayHeight);

	vmacimg = gdk_image_new_bitmap (gtk_widget_get_visual(drawing),
				image_Mem1,
                                macDisplayWidth,
                                macDisplayHeight);

	gtk_widget_show (window);

	vmacgc = gdk_gc_new(drawing->window);
  	fg.pixel = 0xFFFFFFFF;
  	bg.pixel = 0x00000000;
  	gdk_gc_set_foreground (vmacgc, &bg);
  	gdk_gc_set_background (vmacgc, &fg);

	return 0;
}

__inline__ void Screen_Draw(void)
{
	char 	*dst =(char *)vmacimg->mem;
	char	*src;
	int	i;

	if (vPage2 == 1)
		(char *)src = (char *)RAM_Xlate(kMain_Buffer);
	else
		(char *)src = (char *)RAM_Xlate(kAlternate_Buffer);

	memcpy(dst, src, macDisplayHeight*macDisplayWidth/8);

	gdk_draw_image(drawing->window, vmacgc, vmacimg, 0, 0, 0, 0,
		macDisplayWidth, macDisplayHeight);

}

void Screen_Reset(void)
{
	;
}

void file_selection (GtkWidget *widget, gpointer data)
{
    	GtkWidget *filew;
         
        filew = gtk_file_selection_new ("File selection");
         
	gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (filew)->ok_button),
                             "clicked", (GtkSignalFunc) destroy, filew );
         
	gtk_signal_connect_object (GTK_OBJECT (GTK_FILE_SELECTION (filew)->cancel_button),
                                    "clicked", (GtkSignalFunc) gtk_widget_destroy,
                                    GTK_OBJECT (filew));
         
	gtk_file_selection_set_filename (GTK_FILE_SELECTION(filew), "*.dsk");
         
	gtk_widget_show(filew);
}

void file_quit (GtkWidget *widget, gpointer data)
{
	GtkWidget *button, *label;

	if (file_window)
		return;

	file_window = gtk_dialog_new ();

	label = gtk_label_new ((char *) data);
	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (file_window)->vbox), label,
                         TRUE, TRUE, 0);
     	gtk_widget_show (label);

	button = gtk_button_new_with_label ("Ok");
        gtk_signal_connect (GTK_OBJECT (button), "pressed",
                             GTK_SIGNAL_FUNC(Quit_vMac) , NULL);
 	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (file_window)->action_area), button,
                         TRUE, TRUE, 0);
     	gtk_widget_show (button);

	button = gtk_button_new_with_label ("Cancel");
	gtk_signal_connect_object (GTK_OBJECT (button),
                                    "clicked", (GtkSignalFunc) gtk_widget_destroy,
                                    GTK_OBJECT (file_window));

	gtk_box_pack_end (GTK_BOX (GTK_DIALOG (file_window)->action_area), button,
                         TRUE, TRUE, 0);
     	gtk_widget_show (button);

     	gtk_widget_show (file_window);

}

void file_reset (GtkWidget *widget, gpointer data)
{
	GtkWidget *button, *label;

	if (file_window)
		return;

	file_window = gtk_dialog_new ();

	label = gtk_label_new ((char *) data);
	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (file_window)->vbox), label,
                         TRUE, TRUE, 0);
     	gtk_widget_show (label);

	button = gtk_button_new_with_label ("Ok");
        gtk_signal_connect (GTK_OBJECT (button), "pressed",
                             GTK_SIGNAL_FUNC(gdk_m68k_reset) , NULL);
 	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (file_window)->action_area), button,
                         TRUE, TRUE, 0);
     	gtk_widget_show (button);

	button = gtk_button_new_with_label ("Cancel");
        gtk_signal_connect (GTK_OBJECT (button), "pressed",
                             GTK_SIGNAL_FUNC(popup_unmap) , NULL);

	gtk_box_pack_end (GTK_BOX (GTK_DIALOG (file_window)->action_area), button,
                         TRUE, TRUE, 0);
     	gtk_widget_show (button);

     	gtk_widget_show (file_window);

}

__inline__ void Process_Screen_Events() {

        int i = 0;

        while (gtk_events_pending() > 0 && i < 10) {
        	gtk_main_iteration();
		i++;
	}

}

// VIA Interface Functions

UBYTE VIA_GORA6 (void) // Main/Alternate Screen Buffer
{
#	ifdef _VIA_Interface_Debug
  printf("VIA ORA6 attempts to be an input\n");
#	endif
  return 0;
}

void  VIA_PORA6 (UBYTE Data)
{
  vPage2 = Data;
}

UBYTE VIA_GORB6 (void) // Video Beam in Display
{
  return 0; // Assume it is
}

void  VIA_PORB6 (UBYTE Data)
{
#	ifdef _VIA_Interface_Debug
  printf("VIA ORB6 attempts to be an output\n");
#	endif
}

#ifdef HELLO

#define NotChanged -1
        
__inline__ void Screen_Draw(void) {   
        
        UWORD  *screen;
        int     x, y;     
        int     bitTest, bitPos;
        int     currentLineChanged;
        UWORD   currentWord;
        UWORD   previousWord;
        int     firstLineChanged, lastLineChanged;
                          
        if (vPage2 == 1)
                screen = (UWORD *) RAM_Xlate(kMain_Buffer);
        else
               	screen = (UWORD *) RAM_Xlate(kAlternate_Buffer);
        
        firstLineChanged = lastLineChanged = NotChanged;
        
        for( y = 0; y < 384; y++ )
        {
                currentLineChanged = 0;
                for( x = 0; x < 512/16; x++ )
                {
                        currentWord = screen[x + (y * 512/16)];
                        previousWord = previousScreen[x + (y * 512/16)];
                        if( currentWord != previousWord )
                        {
                                bitTest = 1 << (16 - 1);
                                for( bitPos = 0; bitPos < 8; bitPos++ )
                                {
                                        if( bitTest & currentWord )
#ifdef BIGENDIAN
            					XPutPixel( offscreenBuffer, (x*16)+bitPos, y, black.pixel );
#else
            					XPutPixel( offscreenBuffer, (x*16)+bitPos+8, y, black.pixel );
#endif
          				else
#ifdef BIGENDIAN
            					XPutPixel( offscreenBuffer, (x*16)+bitPos, y, white.pixel );
#else
            					XPutPixel( offscreenBuffer, (x*16)+bitPos+8, y, white.pixel );
#endif
          				bitTest = bitTest >> 1;
          			}
        			bitTest = 1 << (8 - 1);
        			for( bitPos = 0; bitPos < 8; bitPos++ )
          			{
          				if( bitTest & currentWord )
#ifdef BIGENDIAN
            					XPutPixel( offscreenBuffer, (x*16)+bitPos+8, y, black.pixel );
#else
            					XPutPixel( offscreenBuffer, (x*16)+bitPos, y, black.pixel );
#endif
          				else
#ifdef BIGENDIAN
            					XPutPixel( offscreenBuffer, (x*16)+bitPos+8, y, white.pixel );
#else
            					XPutPixel( offscreenBuffer, (x*16)+bitPos, y, white.pixel );
#endif
          				bitTest = bitTest >> 1;
                                }
                                currentLineChanged = 1;
                        }
                }

        if( currentLineChanged == 1 ) 
	{
                if( firstLineChanged == NotChanged )
                        firstLineChanged = lastLineChanged = y;
                else
                        lastLineChanged = y;
        }
	}
                                  
        if( firstLineChanged != NotChanged )
        {
                memcpy( previousScreen, screen, (512/8)*384 );
#if MITSHM
                XShmPutImage( display, vmacwin, vmacgc, offscreenBuffer, 0,
                                firstLineChanged, 0, firstLineChanged, 512,
                                lastLineChanged - firstLineChanged + 1, 0);
#else
                XPutImage(display, vmacwin, vmacgc, offscreenBuffer, 0, firstLineChanged,
                                0, firstLineChanged, 512, lastLineChanged - firstLineChanged + 1);
#endif
        }
                
}

#endif
