
/* vMac - Screen-X.c
 * X Window Interface
 * Copyright 1997 Michael Hanni (mhanni@sprintmail.com)
 */

#include "sysconfig.h"
#include "sysdeps.h"
#include "config.h"

#include "memory.h"
#include "Screen.h"
#include "RAM.h"  
#include "VIA.h"
#include "Keyboard.h"
#include "VIA_Interrupts.h"
#include "VIA_Interface.h"
#include "RTC.h"
#include "Native.h"
#include "Sound.h"
#include "Mouse.h"
#include "Sony.h"   
#include "prefs.h"
#include "debug.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>
#include <stdio.h>

#if SCAN
#if MITSHM
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#endif
#endif

UBYTE vPage2 = 1;

Display 	*display;
static int 	screen;
Window 		rootwin, vmacwin;
XColor 		black, white;
GC 		vmacgc;
Visual 		*Xvisual;
int		Xdepth, bit_unit;
Colormap	Xcmap;

int 		macDisplayHeight = 342;
int 		macDisplayWidth = 512;
int		lastmx = 0, lastmy = 0;
int 		buttonstate[3];

extern 		UBYTE KeyCode_Table [128];

#if SCAN
#if MITSHM
XShmSegmentInfo shminfo;
#endif
char            *pixmapBuffer;
UWORD           *previousScreen;
XImage          *offscreenBuffer;
#else
XImage          *vmacimg;
UWORD           *image_Mem1;
#endif

int	inwindow;
Cursor 	blankCursor;

void Quit_vMac(void);

int Screen_Init(int argc, char **argv)
{

	int			i, j;
	XTextProperty 		iconName, windowName;
    	XSetWindowAttributes 	wattr;
    	XPixmapFormatValues 	*xpfvs;
    	XWMHints 		win_hints;
    	XClassHint 		class_hints;
	XGCValues		gcv;
	XSizeHints		size_hints;

    	char 	*display_name = NULL;
    	char 	*window_name = "vMac";
    	char 	*icon_name = "vMac";

	/* Get connection to X Server */

    	if ( (display=XOpenDisplay(display_name)) == NULL )
    	{
         	fprintf(stderr, "Cannot connect to X server.\n");
		Quit_vMac();
    	}
    	screen = DefaultScreen(display);

	Xdepth  = DefaultDepth(display, screen);
  	Xvisual = DefaultVisual(display, screen);
  	Xcmap   = DefaultColormap(display, screen);

	XParseColor(display, Xcmap, "#000000", &black);
		if (!XAllocColor(display, Xcmap, &black))
		fprintf(stderr, "Whoops??\n");
	XParseColor(display, Xcmap, "#ffffff", &white);
		if (!XAllocColor(display, Xcmap, &white))
		fprintf(stderr, "Whoops??\n");
	                                
    	xpfvs = XListPixmapFormats(display, &i);
    	for (j = 0; j < i && xpfvs->depth != Xdepth; j++, xpfvs++);
    	if (j == i) {
        	fprintf(stderr, "Your X server is feeling ill.\n");
        	return 0;
    	}
   	bit_unit = xpfvs->bits_per_pixel;
    	fprintf(stderr, "Using %d bit visual, %d bits per pixel.\n", Xdepth, bit_unit);

    	vmacwin = XCreateSimpleWindow(display, RootWindow(display, rootwin),
		0, 0, macDisplayWidth, macDisplayHeight, 4, 
		white.pixel,
		black.pixel);

	/* Set up some XString properties for window and icon names. */

    	if (XStringListToTextProperty(&window_name, 1, &windowName) ==0) {
      		printf("vMac was unable to allocate a Window name.\n");
      		exit(-1);
    	}

    	if (XStringListToTextProperty(&icon_name, 1, &iconName) ==0) {
      		printf("vMac was unable to allocate an Icon name.\n");
      		exit(-1);
    	}

	/* Hints, hints, hints */

	size_hints.flags = PSize | PPosition | PMinSize | PMaxSize ;
	size_hints.min_width = macDisplayWidth;
	size_hints.max_width = macDisplayWidth;
	size_hints.min_height = macDisplayHeight;
	size_hints.max_height = macDisplayHeight;

    	win_hints.initial_state = NormalState;
    	win_hints.input = True;
    	win_hints.flags = StateHint | WindowGroupHint;
    	win_hints.window_group = vmacwin;

    	class_hints.res_name = "vMac";
    	class_hints.res_class = "vMac";

	/* Is this overkill on XSelectInput? */

    	XSelectInput(display, vmacwin, ExposureMask | KeyPressMask | KeyReleaseMask |
      		ButtonPressMask | ButtonReleaseMask | StructureNotifyMask | PointerMotionMask |
      		EnterWindowMask | LeaveWindowMask );

    	vmacgc = XCreateGC(display, vmacwin, 0, NULL);
	XSetState(display, vmacgc, black.pixel, white.pixel, GXcopy, AllPlanes);

	XSetWMProperties(display, vmacwin, &windowName, &iconName, argv, argc,
			&size_hints, &win_hints, &class_hints);

	buttonstate[0] = buttonstate[1] = buttonstate[2] = 0;
	
	if (thePrefs.NoXmouse == 1)
	{
		blankCursor = XCreatePixmapCursor(display, XCreatePixmap(display, vmacwin, 1, 1, 1),
                                   	   XCreatePixmap(display, vmacwin, 1, 1, 1),
                                           &black, &white, 0, 0);
		XDefineCursor(display, vmacwin, blankCursor);
	}

#if SCAN        
        previousScreen = calloc(1,(512/8)*384 ); /* Ugh. */
#if MITSHM
        fprintf(stderr, "vMac: Using MIT-SHaredMemory Extensions.\n");
        offscreenBuffer = XShmCreateImage(display, Xvisual, Xdepth, ZPixmap,
        NULL, &shminfo, 512, 384);
        shminfo.shmid = shmget(IPC_PRIVATE, 4*512*384, IPC_CREAT|0777);
        shminfo.shmaddr = offscreenBuffer->data = (char *)shmat(shminfo.shmid,
        0, 0);
        shminfo.readOnly = False;
        XShmAttach(display, &shminfo);
        XSync(display, 0);
#else   
        pixmapBuffer = malloc( 4*512*384 ); /* Ugh. Fix */
        offscreenBuffer = XCreateImage( display, Xvisual, Xdepth, ZPixmap, 0,
                (char *) pixmapBuffer, 512, 384, (Xdepth==8)?8:32, 0);
#endif
#else
        image_Mem1 = (UWORD *)calloc(1, 64 * macDisplayHeight);
        vmacimg = XCreateImage(display, Xvisual, 1, XYBitmap, 0, (char *)image_Mem1,
                macDisplayWidth, macDisplayHeight, 32, 512 / 8);
        vmacimg->bitmap_bit_order = MSBFirst;
        vmacimg->byte_order = MSBFirst;
#endif  

    	XMapWindow(display, vmacwin);

    	XSync(display, 0);

    	return 0;
}

#if SCAN
        
#define NotChanged -1
        
__inline__ void Screen_Draw(void) {   
        
        UWORD  *screen;
        int     x, y;     
        int     bitTest, bitPos;
        int     currentLineChanged;
        UWORD   currentWord;
        UWORD   previousWord;
        int     firstLineChanged, lastLineChanged;
                          
        if (vPage2 == 1)
                screen = (UWORD *) RAM_Xlate(kMain_Buffer);
        else
               	screen = (UWORD *) RAM_Xlate(kAlternate_Buffer);
        
        firstLineChanged = lastLineChanged = NotChanged;
        
        for( y = 0; y < 384; y++ )
        {
                currentLineChanged = 0;
                for( x = 0; x < 512/16; x++ )
                {
                        currentWord = screen[x + (y * 512/16)];
                        previousWord = previousScreen[x + (y * 512/16)];
                        if( currentWord != previousWord )
                        {
                                bitTest = 1 << (16 - 1);
                                for( bitPos = 0; bitPos < 8; bitPos++ )
                                {
                                        if( bitTest & currentWord )
#ifdef BIGENDIAN
            					XPutPixel( offscreenBuffer, (x*16)+bitPos, y, black.pixel );
#else
            					XPutPixel( offscreenBuffer, (x*16)+bitPos+8, y, black.pixel );
#endif
          				else
#ifdef BIGENDIAN
            					XPutPixel( offscreenBuffer, (x*16)+bitPos, y, white.pixel );
#else
            					XPutPixel( offscreenBuffer, (x*16)+bitPos+8, y, white.pixel );
#endif
          				bitTest = bitTest >> 1;
          			}
        			bitTest = 1 << (8 - 1);
        			for( bitPos = 0; bitPos < 8; bitPos++ )
          			{
          				if( bitTest & currentWord )
#ifdef BIGENDIAN
            					XPutPixel( offscreenBuffer, (x*16)+bitPos+8, y, black.pixel );
#else
            					XPutPixel( offscreenBuffer, (x*16)+bitPos, y, black.pixel );
#endif
          				else
#ifdef BIGENDIAN
            					XPutPixel( offscreenBuffer, (x*16)+bitPos+8, y, white.pixel );
#else
            					XPutPixel( offscreenBuffer, (x*16)+bitPos, y, white.pixel );
#endif
          				bitTest = bitTest >> 1;
                                }
                                currentLineChanged = 1;
                        }
                }

        if( currentLineChanged == 1 ) 
	{
                if( firstLineChanged == NotChanged )
                        firstLineChanged = lastLineChanged = y;
                else
                        lastLineChanged = y;
        }
	}
                                  
        if( firstLineChanged != NotChanged )
        {
                memcpy( previousScreen, screen, (512/8)*384 );
#if MITSHM
                XShmPutImage( display, vmacwin, vmacgc, offscreenBuffer, 0,
                                firstLineChanged, 0, firstLineChanged, 512,
                                lastLineChanged - firstLineChanged + 1, 0);
#else
                XPutImage(display, vmacwin, vmacgc, offscreenBuffer, 0, firstLineChanged,
                                0, firstLineChanged, 512, lastLineChanged - firstLineChanged + 1);
#endif
        }
                
}

#else

__inline__ void Screen_Draw(void)
{
        UWORD   *dst =(UWORD *)vmacimg->data;
        UWORD  	*src;

        if (vPage2 == 1)
                src = (UWORD *)RAM_Xlate(kMain_Buffer);
        else
               	src = (UWORD *)RAM_Xlate(kAlternate_Buffer);

	memcpy(dst, src, macDisplayHeight*macDisplayWidth/8);

        XPutImage(display, vmacwin, vmacgc, vmacimg, 0, 0, 0, 0,
                  vmacimg->width, vmacimg->height);

}
#endif

void Screen_Reset(void)
{
	;
}

__inline__ void Process_Screen_Events(void)
{
   int i = 0, j;
   XEvent event;  
   KeySym Key, Key2 = 0x01C, Key3 = 0x01D, Key4 = 0x01E, Key5 = 0x01F;
   char ourKey;
   UWORD Modifier;
   XKeyEvent *xkey;

   while(XPending(display) > 0)
   {

     XNextEvent(display, &event);

	switch(event.type) {
         	case KeyPress:
			xkey = (XKeyEvent *)&event;
			j = XLookupString(xkey, &ourKey, 1, &Key, NULL);
			switch (Key) {
				case XK_F7:
      					Call_ATrap(0xA94C);
					break;
				case XK_F12:
					Quit_vMac();
					break;
				case XK_Left:
					Keyboard_Down(Key2);
					break;
				case XK_Right:
					Keyboard_Down(Key3);
					break;
				case XK_Up:
					Keyboard_Down(Key4);
					break;
				case XK_Down:
					Keyboard_Down(Key5);
					break;
                		case XK_Meta_L:
                		case XK_Meta_R:
                		case XK_Alt_L:
                		case XK_Alt_R:
				case XK_Control_L:
				case XK_Control_R:
				case XK_Shift_L:
				case XK_Shift_R:
				case XK_Caps_Lock:
					break;
				default:
					if (xkey->state & ShiftMask)
						Modifier = 0x0200;
					else if (xkey->state & ControlMask)
						Modifier = 0x0800;
                        		else if (xkey->state & Mod1Mask)
                                		Modifier = 0x0100;
					else if (xkey->state & LockMask)
						Modifier = 0x0400;
					else
						Modifier = 0x0000;
					if (thePrefs.Kanji == 1) {
						Key &= 0xff;
                        			if ( Key==0xff) Key=0x08; // JIS-Key Only?
                        			Key=Key | (KeyCode_Table[Key]<<8)  ;
                        			if (Key==0x490d) Key=0x240d;
					}
					PC_PostEvent (3, Key, Modifier);
					break;
			}
			break;
         	case KeyRelease:
			xkey = (XKeyEvent *)&event;
			j = XLookupString(xkey, &ourKey, 1, &Key, NULL);
			switch (Key) {
				case XK_Left:
					Keyboard_Up(Key2);
					break;
				case XK_Right:	
					Keyboard_Up(Key3);
					break;
				case XK_Up:
					Keyboard_Up(Key4);
					break;
				case XK_Down:
					Keyboard_Up(Key5);
					break;
                		case XK_Meta_L:
                		case XK_Meta_R:
                		case XK_Alt_L:
                		case XK_Alt_R:
				case XK_Control_L:
				case XK_Control_R:
				case XK_Shift_L:
				case XK_Shift_R:
				case XK_Caps_Lock:
					break;
				default:
					if (xkey->state & ShiftMask)
						Modifier = 0x0200;
					else if (xkey->state & ControlMask)
						Modifier = 0x0800;
                        		else if (xkey->state & Mod1Mask)
                                		Modifier = 0x0100;
					else if (xkey->state & LockMask)
						Modifier = 0x0400;
					else
						Modifier = 0x0000;
					if (thePrefs.Kanji == 1) {
						Key &= 0xff;
                        			if ( Key==0xff) Key=0x08; // JIS-Key Only?
                        			Key=Key | (KeyCode_Table[Key]<<8)  ;
                        			if (Key==0x490d) Key=0x240d;
					}
					PC_PostEvent (4, Key, Modifier);
					break;
			}
			break;
         	case ButtonPress:
            		buttonstate[((XButtonEvent *)&event)->button-1] = 1;
                     	break;
        	case ButtonRelease:
            		buttonstate[((XButtonEvent *)&event)->button-1] = 0;
			break;
         	case EnterNotify:
			lastmx = ((XCrossingEvent *)&event)->x;
			lastmy = ((XCrossingEvent *)&event)->y;
            		inwindow = 1;
			break;
         	case LeaveNotify:
			inwindow = 0;
			break;
         	case MotionNotify:
            		if (inwindow) {
                		lastmx = ((XMotionEvent *)&event)->x;
                		lastmy = ((XMotionEvent *)&event)->y;
            		}
			break;
		default:
			break;
	}
   }
}

// VIA Interface Functions

UBYTE VIA_GORA6 (void) // Main/Alternate Screen Buffer
{
#	ifdef _VIA_Interface_Debug
  printf("VIA ORA6 attempts to be an input\n");
#	endif
  return 0;
}

void  VIA_PORA6 (UBYTE Data)
{
  vPage2 = Data;
}

UBYTE VIA_GORB6 (void) // Video Beam in Display
{
  return 0; // Assume it is
}

void  VIA_PORB6 (UBYTE Data)
{
#	ifdef _VIA_Interface_Debug
  printf("VIA ORB6 attempts to be an output\n");
#	endif
}

