 /* 
  * Fatal exceptions handler [including m68k critical exceptions]
  *
  * (c) Yoav Shadmi, 1998
  */

#include "sysconfig.h"
#include "sysdeps.h"

#include <signal.h>
#include <ctype.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include "config.h"
#include "options.h"
#include "debug.h"
#include "memory.h"
#include "custom.h"
#include "reentrant_newcpu.h"

void Quit_vMac(void);
void vMac_Install_SignalHandler(void);
void vMac_Uninstall_SignalHandler(void);

void fatal_error (char *_sig)
{
	fprintf(stderr, "vMac: Fatal Exception: %s\n", _sig);

  	Quit_vMac();
}


static int caught_fatal_sig = 0;

static RETSIGTYPE on_signal (int sig_num)
{
  	char *sig;

  	if (caught_fatal_sig)
    		kill (getpid (), sig_num);
  	caught_fatal_sig = 1;

  	switch (sig_num)
    	{
      		case SIGHUP : sig = "SIGHUP  [modem hangup].\nvMac cannot continue and will be terminated."; break;    
      		case SIGQUIT: sig = "SIGQUIT [quit() requested].\nvMac cannot continue and will be terminated."; break;
      		case SIGABRT: sig = "SIGABRT [abort() requested].\nvMac cannot continue and will be terminated."; break;
      		case SIGSEGV: sig = "SIGSEGV [general protection fault].\nvMac cannot continue and will be terminated."; break;
      		case SIGPIPE: sig = "SIGPIPE [broken pipe] - restart your system."; break;
      		case SIGTERM: sig = "SIGTERM [termination requested].\nvMac cannot continue and will be terminated."; break;
      		case SIGFPE : sig = "SIGFPE  [FPU fatal exception].\nvMac cannot continue and will be terminated."; break;
      		case SIGKILL: sig = "SIGKILL. vMac cannot continue and will be terminated."; break;
      		case SIGUSR2: sig = "The emulated m68k has encountered an illegal instruction.vMac cannot continue and will be terminated."; break;
      		case SIGINT:  sig = "SIGINT [interrupt].\nvMac has encountered a user interrupt and will quit."; break;
      		case SIGBUS:  sig = "SIGBUS [bus error].\nvMac cannot continue and will be terminated."; break;
      		default:      sig = "Unknown exception.\nvMac cannot continue and will be terminated."; break;
    	}
  	fatal_error (sig);
}

void vMac_Install_SignalHandler(void) {

	/* Handle some signals */
  	signal (SIGHUP,  on_signal);
  	signal (SIGQUIT, on_signal);
  	signal (SIGABRT, on_signal);
  	signal (SIGSEGV, on_signal);
  	signal (SIGPIPE, on_signal);
  	signal (SIGTERM, on_signal);
   	signal (SIGFPE,  on_signal);
   	signal (SIGKILL, on_signal);
   	signal (SIGUSR2, on_signal);
  	signal (SIGINT,  on_signal);
  	signal (SIGBUS,  on_signal);

	printf("vMac: vMac_SignalHandler() Installed.\n");

}

void vMac_Uninstall_SignalHandler(void) {
                        
   	signal(SIGHUP,  SIG_DFL);
   	signal(SIGQUIT, SIG_DFL);
   	signal(SIGABRT, SIG_DFL);
   	signal(SIGSEGV, SIG_DFL);
   	signal(SIGPIPE, SIG_DFL);
   	signal(SIGTERM, SIG_DFL);
   	signal(SIGFPE,  SIG_DFL);
   	signal(SIGKILL, SIG_DFL);
   	signal(SIGUSR1, SIG_DFL);
   	signal(SIGUSR2, SIG_DFL);
   	signal(SIGINT,  SIG_DFL);
   	signal(SIGBUS,  SIG_DFL);

        printf("vMac: vMac_SignalHandler() Uninstalled.\n");
                        
}
