/* Support for Gemulator ROM Board */

/*
 * Much of this code was written by Darek Michocka,
 * and adapted to work in vMac by Weston, and ported
 * to Linux by Michael Samuel.
 *
 * If this code fails, try adding defining SLOW_IO, and see how that works.
 */

#define SLOW_IO

#include <stdio.h>
#include <asm/io.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/io.h>

/* Define TEST to make a linkable ROM dumper */
#ifdef TEST
#define SOCKNUM 1
#define NUMSOCKS 2

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#endif

#include "sysconfig.h"
#include "sysdeps.h"
#include "gemulator.h"

/*
 * Change this if your board is on a different I/O Port
 * This is statically compiled into the binary *deliberately*
 */
#define GEM_BASE 0x240

/*
 * Much of this code was written by Darek Michocka,
 * and adapted to work in vMac by Weston, and ported
 * to Linux, and thrown into BasiliskII by Michael Samuel
 */

/*
 * Don't forget to drop root permissions after trying to load the ROM,
 * especially before trying to read a ROM from a file...
 */

#define cbSocket (128*1024)
#define countSocket (8)

unsigned char bigbuf[cbSocket*countSocket*2];

int Gemulator_LoadROM(UBYTE * NewROM, UWORD Base, UBYTE Socket, UBYTE Length)
{
	int i;
	int rgsize[countSocket];

	--Socket; //Our socket 1 is really socket 0
	
	/* Give I/O Permissions to the Base address */
	if(ioperm(GEM_BASE, 1, 1) == -1) {
		fprintf(stderr, "Cannot get IO permissions: %s\n", 
				strerror(errno));
		return 0;
	}

#ifdef SLOW_IO
	if(ioperm(0x80, 1, 1) == -1) {
		fprintf(stderr, "Cannot get IO permissions: %s\n",
				strerror(errno));
		return 0;
	}
	outb_p(0, GEM_BASE);
#else
	outb(0, GEM_BASE);
#endif
	for (i = 0; i < (Socket+Length)*cbSocket; i++) {
#ifdef SLOW_IO
		bigbuf[i] = inb_p(GEM_BASE);
#else
		bigbuf[i] = inb(GEM_BASE);
#endif
	}

	/* Drop I/O permissions, now that we're done */
	ioperm(GEM_BASE, 1, 0);
#ifdef SLOW_IO
	ioperm(0x80, 1, 0);
#endif
	for (i = 0; i < countSocket; i++) {
		unsigned char *pb = &bigbuf[i * cbSocket];
		int size = cbSocket/1024;
		while ((size > 32) && !memcmp(pb, pb+(size*512), size*512)) {
			size /= 2;
		}
		rgsize[i] = size;
	}

	// scramble the bytes from socket order to byte order
	memmove(bigbuf + countSocket * cbSocket, bigbuf + Socket * cbSocket,
			Length * cbSocket);

	for (i = 0; i < countSocket * cbSocket; i++) {
	        bigbuf[i] = bigbuf[(countSocket + (i%Length)) *
				cbSocket + (i/Length)];
	}
	memcpy(NewROM, bigbuf, 1024 * rgsize[Socket] * Length);

	return(1024 * rgsize[Socket] * Length); //Return size
}

#ifdef TEST
int main(int argc, char **argv)
{
	unsigned char rombuff[(1024*128*NUMSOCKS)];
	int fd;
	int size = Gemulator_LoadROM(rombuff, SOCKNUM, NUMSOCKS);

	if(size == 0) {
		printf("ROM Loading Failed\n");
		exit(1);
	}
	printf("ROM Size: %d bytes\n", size);
	fd = open("Gemulator_dump", O_CREAT | O_WRONLY);
	if(fd == -1) {
		fprintf(stderr, "Cannot dump ROM file\n");
		exit(1);
	}
	if(write(fd, rombuff, (ssize_t)size) != (ssize_t)size) {
		fprintf(stderr, "error writing file\n");
		close(fd);
		exit(1);
	}
	close(fd);
	return 0;
}
#endif
	
